eval '(exit $?0)' && eval 'exec perl -S $0 ${1+"$@"}'
  & eval 'exec perl -S $0 $argv:q'
  if 0;

#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2009 - 2012.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#
##############################################################################

=head1 NAME

cqtsadmin.pl

=head1 SYNOPSIS

B<cqtsadmin.pl> B<[-man]> B<[-help]>

=head1 OPTIONS

=over 4

=item -help

Print a brief help message.

=item -man

Print the man page.

=back

=head1 DESCRIPTION

=head1 AUTHOR

=cut


use CQPerlExt;

use CQTsAdmin::L10N;

use strict;
use warnings;
use IO::File;
use XML::Parser;
use HTML::Entities;
use File::Copy;
use File::Find;
use File::Path;
use File::Spec;
use Getopt::Long;
use Sys::Hostname;
use Data::Dumper;
use Encode;


my $langMsgText = CQTsAdmin::L10N->get_handle()
  || die "Cannot find an acceptable language module!?!";

# replace PERL's print function so it can log as well
my $logFileHandler;				# the file handle to the log file
my $cqtsadminLogFile = "";		# the log file name
my $logDataBacklog = "";		# holds log data in memory till when the log file is ready

sub printAndLog;
sub openLogger;
sub closeLogger;


require "CQPerlExt.pm" || die;

# get localized message strings
my $stat_START = $langMsgText->maketext("START_MSG_TEXT");
my $stat_DONE = $langMsgText->maketext("DONE_MSG_TEXT");

my $msgLabel_INFO = $langMsgText->maketext("INFO_MSG_TEXT");
my $msgLabel_WARN = $langMsgText->maketext("WARN_MSG_TEXT");
my $msgLable_ERROR = $langMsgText->maketext("ERROR_MSG_TEXT");
my $msgLable_oS_ERROR = $langMsgText->maketext("ERROR_MSG_TEXT_OS");


my $gAboutThisFTSFile = "";

my $gDieKind_must_match = "must-match";
my $gDieKind_must_not_match = "must-not-match";
my $gDieKind_none = "none";
my $gTerminationErrorMsg = $msgLable_ERROR."*************** ".$langMsgText->maketext("TERMINATION_MSG")." ****************";
my $gMaxWASProfileLen = length("123456789012345678901234567");

my $gExternalCmdMarker = $ENV{"CQ_FTS_XCMD_MARKER"};						# Not documented
my $gSuppressExternalCmdOutput = $ENV{"CQ_FTS_SUPPRESS_XCMD_OUTPUT"};   	# Not documented, must be set to "yes" to take effect

my $gCqFtsEncoding = $ENV{"CQ_FTS_ENCODING"};
if (!defined($gCqFtsEncoding))
{
	$gCqFtsEncoding = $langMsgText->maketext("CODEPAGE_ENCODING");
}

my $gIsReplicated = 0;
my $gWinServiceStartUpType = "";
my $gBackupDestinationDir = "";
my $gScrubOplogBeforeDate = "";
my $gPrepUpgdFtsWASProfiles = "";
my $gRestoreFtsWASProfiles = "";
my $gShowScenarios = "";
my $gShowScenariosOption = 0;

my @gCmdLineArgvList;
my @gCmdToExecute;
my %gDispOrdFieldList = ();

my $gTopLevelSate = "";
my %gFtsScenarios = ();

my $gVerboseOutput = 1;		# for this release, we are always going to be verbose

my $optUserName = "";						# --username=s						[takes a string]
my $optPassword = "";						# --password=s						[takes a string]
my $optDbSet = "";							# --dbset=s							[takes a string]
my $optuserDB = "";							# --userdb=s						[takes a string]
my $optCqFtsHomeDir = "";					# --ftshome=s						[takes a string]
my $optInitCqFTS = "";						# --init_cq_fts
my $optSetupCqFTS = "";						# --setup_cq_fts
my $optCopyFTSTemplate = "";				# --copy_fts_template
my $optGenerateEntityFile = "";				# --gen_entity_file
my $optGenerateFTSFiles = "";				# --gen_fts_files
my $optCustomizeCQSearchPropertyFile = "";	# --cust_fts_files
my $optCustomizeSolrFile = "";				# --cust_solr_files
my $optStopWASSearchProfile = "";			# --stop_fts_was_profile
my $optSetSolrHome = "";					# --set_solr_home
my $optStartWASSearchProfile = "";			# --start_fts_was_profile
my $optRunInstallUtil = "";					# --enable_cqweb_fts
my $optRunBatchIndex = "";					# --run_batch_idx
my $optEnableUpdateIndex = "";				# --start_update_idx
my $optUpdateGenFiles = "";					# --update_fts_prop_files
my $optDisableFTS = "";						# --disable_cqweb_fts
my $optSetWASMaxMem = "";					# --set_was_max_mem
my $optStopUpdateIndex = "";				# --stop_update_idx
my $optFreshRunBatchIndex = "";				# --fresh_batch_idx
my $optFreshRunUpdateIndex = "";			# --fresh_update_idx
my $optAddRecordType = ""; 					# --add_record_type
my $optRemoveRecordType = "";				# --remove_record_type
my $optCreateCqFtsWasProfile = "";			# --create_fts_was_profile=s		[takes a string]
my $optDeleteCqFtsWasProfile = "";			# --delete_fts_was_profile
my $optOptimizeIndex = "";					# --optimize_idx
my $optRemoveLuceneIdxLock = "";			# --remove_lucene_idx_lock
my $optRepairRecords = "";					# --repair_records
my $optScrubOplog = "";						# --scrub_oplog
my $optBackupCqFTS = "";                    # --backup_fts=s					[takes a string]
my $optArchiveCqFTS = "";					# --archive_fts
my $optClearState = "";						# --clear_state
my $optLockCqFts = "";						# --lock_cq_fts
my $optUnlockCqFts = "";					# --unlock_cq_fts
my $optValidateCqFtsDeplo = "";				# --validate_cq_fts_deplo
my $optPrepUpgdFtsWASProfiles = "";			# --prep_upgd_was_profiles=s		[takes a string]
my $optRestoreFtsWASProfiles = "";			# --restore_was_profiles=s			[takes a string]
my $optGatherDiagnosticData = "";			# --gather_diagnostic_data
my $optShowScenarios = "";					# --show_scenarios=s				[takes a string]
my $optVerbose = "";						# --verbose
my $optHelp = "";							# --help

# get the order of the comand line options
foreach my $argnum (0 .. $#ARGV)
{
	if ($ARGV[$argnum] =~ m/^-/)
	{
		my $tmp = $ARGV[$argnum];
		$tmp =~ s/^-//;
		$tmp =~ s/^-//;
		
		if ($tmp eq "show_scenarios")
		{
			$gShowScenariosOption = 1;
			next;
		}
		if ($tmp eq "verbose")
		{
			next;
		}
		
		push(@gCmdLineArgvList, $tmp);
	}
}

# Get flag values
my $result = GetOptions(
						 "username=s" => \$optUserName, 
						 "password=s" => \$optPassword, 
						 "dbset=s" => \$optDbSet, 
						 "userdb=s" => \$optuserDB, 
						 "ftshome=s" => \$optCqFtsHomeDir, 
						 
						 "init_cq_fts" => \$optInitCqFTS,
						 "setup_cq_fts" => \$optSetupCqFTS, 
						 
						 "copy_fts_template" => \$optCopyFTSTemplate, 
						 "gen_entity_file" => \$optGenerateEntityFile, 
						 
						 "gen_fts_files" => \$optGenerateFTSFiles, 
						 "cust_fts_files" => \$optCustomizeCQSearchPropertyFile, 
						 "cust_solr_files" => \$optCustomizeSolrFile, 
						 "stop_fts_was_profile" => \$optStopWASSearchProfile, 
						 "set_solr_home" => \$optSetSolrHome, 
						 "start_fts_was_profile" => \$optStartWASSearchProfile, 
						 "enable_cqweb_fts" => \$optRunInstallUtil, 
						 "run_batch_idx" => \$optRunBatchIndex, 
						 "start_update_idx" => \$optEnableUpdateIndex, 
						 
						 "update_fts_prop_files" => \$optUpdateGenFiles, 
						 "disable_cqweb_fts" => \$optDisableFTS, 
						 "set_was_max_mem" => \$optSetWASMaxMem, 
						 "stop_update_idx" => \$optStopUpdateIndex, 
						 "fresh_batch_idx" => \$optFreshRunBatchIndex, 
						 "fresh_update_idx" => \$optFreshRunUpdateIndex, 
						 
						 "add_record_type" => \$optAddRecordType, 	# *
						 "remove_record_type" => \$optRemoveRecordType, 	# *
						 
						 "create_fts_was_profile=s" => \$optCreateCqFtsWasProfile, 
						 "delete_fts_was_profile" => \$optDeleteCqFtsWasProfile, 
						 
						 "optimize_idx" => \$optOptimizeIndex, 
						 "remove_lucene_idx_lock" => \$optRemoveLuceneIdxLock, 
						 "repair_records" => \$optRepairRecords, 	# *
						 "scrub_oplog=s" => \$optScrubOplog,
						 "backup_fts=s" => \$optBackupCqFTS, 
						 "archive_fts" => \$optArchiveCqFTS, 
						 
						 "clear_state" => \$optClearState,
						 
						 "lock_cq_fts" => \$optLockCqFts, 
						 "unlock_cq_fts" => \$optUnlockCqFts, 
						 
						 "validate_cq_fts_deplo" => \$optValidateCqFtsDeplo, 
						 
						 "prep_upgd_was_profiles=s" => \$optPrepUpgdFtsWASProfiles, 
						 "restore_was_profiles=s" => \$optRestoreFtsWASProfiles, 
						 
						 "gather_diagnostic_data" => \$optGatherDiagnosticData, 
						 
						 "show_scenarios:s" => \$optShowScenarios, 
						 
						 "verbose" => \$optVerbose, 
						 
						 "help" => \$optHelp
						);


# they are asking for help, show it and then exit
if ($optHelp eq "1")
{
	Usage("", 1);
}

# they are asking for scenarios, show it and then exit
if (($optShowScenarios ne "") or ($gShowScenariosOption eq 1))
{
	$gShowScenarios = $optShowScenarios;
	$optShowScenarios = "1";
	Func_show_scenarios();
	exit 1;
}

# do they want verbose output?
if ($optVerbose ne "")
{
	Func_verbose();
	$optVerbose = "";
}


# verify they used proper command line options (i.e.: they didn't just pass in no option(s))
if ($result eq "")
{
	my $dieMsg1 = $langMsgText->maketext("INVALID_COMMAND_LINE_OPTION");
	my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
	my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
	dealWithError($dieMsg, 1);
}
else
{
	my $haveRequiredOptions = 0;
	my $haveAtLeastOneOption = 0;
	
	# make sure the min. required command-line options and at least one additional option is used
	foreach my $item (@gCmdLineArgvList)
	{
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome"))
		{
			$haveRequiredOptions++;
		}
		else
		{
			$haveAtLeastOneOption++;
		}
	}
	
	# are all required command-line options given?
	if ($haveRequiredOptions ne 5)
	{
		my $dieMsg1 = $langMsgText->maketext("MISSING_REQUIRED_OPTIONS");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
	
	# is at least one additional command-line option given?
	if ($haveAtLeastOneOption eq 0)
	{
		my $dieMsg1 = $langMsgText->maketext("NO_OPERATIONAL_COMMAND_OPTION");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
}

# deal with some exceptions with the command line option:
# if init_cq_fts, or setup_cq_fts is used, nothing else is allowed
if (($optInitCqFTS eq "1") or ($optSetupCqFTS eq "1"))
{
	foreach my $item (@gCmdLineArgvList)
	{
		my $foundAt = index($item, '=');	# if the command takes arg, remove the the arg part

		if ($foundAt ne -1)
		{
			$item = substr($item, 0, $foundAt);
		}
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome") or ($item eq "init_cq_fts") or
			($item eq "setup_cq_fts") or ($item eq "create_fts_was_profile"))
		{
			next;
		}
		else
		{
			my $dieMsg1 = $langMsgText->maketext("NOT_SUPPORTED_COMMAND_OPTION");
			my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
			my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
			dealWithError($dieMsg, 1);
		}
	}
}

# deal with command line options that accept parameters
if ($optCreateCqFtsWasProfile ne "")
{
	# deal with: --create_fts_was_profile
	$optCreateCqFtsWasProfile = lc($optCreateCqFtsWasProfile);
	if (($optCreateCqFtsWasProfile eq "manual") or 
		($optCreateCqFtsWasProfile eq "automatic") or 
		($optCreateCqFtsWasProfile eq "disabled"))
	{
		$gWinServiceStartUpType = $optCreateCqFtsWasProfile;
		$optCreateCqFtsWasProfile = "1";
	}
	else
	{
		my $dieMsg1 = $langMsgText->maketext("INVALID_CREATE_PROFILE");
		my $dieMsg2 = $langMsgText->maketext("USE_HELP_MSG");
		my $dieMsg = "\n".$dieMsg1."\n"."\n".$dieMsg2."\n";
		dealWithError($dieMsg, 1);
	}
}
if ($optBackupCqFTS ne "")
{
	$gBackupDestinationDir = $optBackupCqFTS;
	$optBackupCqFTS = "1";
}
if ($optScrubOplog ne "")
{
	$gScrubOplogBeforeDate = $optScrubOplog;
	$optScrubOplog = "1";
}
if ($optPrepUpgdFtsWASProfiles ne "")
{
	$gPrepUpgdFtsWASProfiles = $optPrepUpgdFtsWASProfiles;
	$optPrepUpgdFtsWASProfiles = "1";
}
if ($optRestoreFtsWASProfiles ne "")
{
	$gRestoreFtsWASProfiles = $optRestoreFtsWASProfiles;
	$optRestoreFtsWASProfiles = "1";
}


# Print command-line value
printAndLog ""."\n";
printAndLog ""."\n";
printAndLog "username ->               : $optUserName"."\n";
printAndLog "password ->               : ********"."\n";
printAndLog "dbset ->                  : $optDbSet"."\n";
printAndLog "userdb ->                 : $optuserDB"."\n";
printAndLog "ftshome ->                : $optCqFtsHomeDir"."\n";
printAndLog "init_cq_fts ->            : $optInitCqFTS"."\n";
printAndLog "setup_cq_fts ->           : $optSetupCqFTS"."\n";
printAndLog "copy_fts_template ->      : $optCopyFTSTemplate"."\n";
printAndLog "gen_entity_file ->        : $optGenerateEntityFile"."\n";
printAndLog "gen_fts_files ->          : $optGenerateFTSFiles"."\n";
printAndLog "cust_fts_files ->         : $optCustomizeCQSearchPropertyFile"."\n";
printAndLog "cust_solr_files ->        : $optCustomizeSolrFile"."\n";
printAndLog "stop_fts_was_profile ->   : $optStopWASSearchProfile"."\n";
printAndLog "set_solr_home ->          : $optSetSolrHome"."\n";
printAndLog "start_fts_was_profile ->  : $optStartWASSearchProfile"."\n";
printAndLog "enable_cqweb_fts ->       : $optRunInstallUtil"."\n";
printAndLog "run_batch_idx ->          : $optRunBatchIndex"."\n";
printAndLog "start_update_idx ->       : $optEnableUpdateIndex"."\n";
printAndLog "update_fts_prop_files ->  : $optUpdateGenFiles"."\n";
printAndLog "disable_cqweb_fts ->      : $optDisableFTS"."\n";
printAndLog "set_was_max_mem ->        : $optSetWASMaxMem"."\n";
printAndLog "stop_update_idx ->        : $optStopUpdateIndex"."\n";
printAndLog "fresh_batch_idx ->        : $optFreshRunBatchIndex"."\n";
printAndLog "fresh_update_idx ->       : $optFreshRunUpdateIndex"."\n";
printAndLog "add_record_type ->        : $optAddRecordType"."\n";
printAndLog "remove_record_type ->     : $optRemoveRecordType"."\n";
printAndLog "create_fts_was_profile -> : $optCreateCqFtsWasProfile"."\n";
printAndLog "delete_fts_was_profile -> : $optDeleteCqFtsWasProfile"."\n";
printAndLog "optimize_idx ->           : $optOptimizeIndex"."\n";
printAndLog "remove_lucene_idx_lock -> : $optRemoveLuceneIdxLock"."\n";
printAndLog "repair_records ->         : $optRepairRecords"."\n";
printAndLog "scrub_oplog ->            : $optScrubOplog"."\n";
printAndLog "backup_fts ->             : $optBackupCqFTS"."\n";
printAndLog "archive_fts ->            : $optArchiveCqFTS"."\n";
printAndLog "clear_state ->            : $optClearState"."\n";
printAndLog "lock_cq_fts ->            : $optLockCqFts"."\n";
printAndLog "unlock_cq_fts ->          : $optUnlockCqFts"."\n";
### ### printAndLog "validate_cq_fts_deplo ->  : $optValidateCqFtsDeplo"."\n";
printAndLog "prep_upgd_was_profiles -> : $optPrepUpgdFtsWASProfiles"."\n";
printAndLog "restore_was_profiles ->   : $optRestoreFtsWASProfiles"."\n";
printAndLog "gather_diagnostic_data -> : $optGatherDiagnosticData"."\n";
printAndLog "show_scenarios ->         : $optShowScenarios"."\n";
### ### printAndLog "verbose ->                : $optVerbose"."\n";
printAndLog "help ->                   : $optHelp"."\n";
printAndLog "**result's ->             : $result"."\n";
printAndLog ""."\n";
printAndLog ""."\n";


# fix '--ftshome'
$optCqFtsHomeDir = File::Spec->rel2abs($optCqFtsHomeDir);
$optCqFtsHomeDir =~ s/\\/\//sg;

if (-d $optCqFtsHomeDir)
{
	# we are given a directory
	if ((substr($optCqFtsHomeDir, -1) ne "/") or (substr($optCqFtsHomeDir, -1) ne "\\"))
	{
		$optCqFtsHomeDir .= "/";
	}
}


my $cqVersion = "";

my %defValues = ();
my %newValues = ();

my $tag          = "";
my $defaultValue = "";
my $newValue     = "";
my $data         = "";

my $isFTSEnabled = 0;
my $isInDefaultValue = 0;
my $isInNewValue = 0;

my $element = "";

sub trim;
sub compareArrays;
sub cleanArrayList;
sub dealWithError;
sub chdirAndValidate;
sub validateFileGenerated;
sub executeExternalCommand;
sub createAboutThisFts;
sub resolveFTSDataDir;
sub setupCqFts;
sub createNewCqFtsSetup;
sub getSolrCopyFieldData;
sub getSolrFieldData;
sub findExistingCqFtsSetup;
sub copyFiles;
sub copyFilesMatch;
sub validateUser;
sub cleanupConfigUserData;
sub fixPathSlashes;
sub findWebXMLFolder;
sub findServerXMLFolder;
sub getHeapSizeAndJVMArg;
sub getEntityFileWithoutDispFieldInfo;
sub buildScenariosFromXMLFile;
sub canDeleteDefaultProfile_cqsearchprofile;

sub Func_copy_fts_template;
sub Func_gen_entity_file;
sub Func_gen_fts_files;
sub Func_cust_fts_files;
sub Func_cust_solr_files;
sub Func_stop_fts_was_profile;
sub Func_set_solr_home;
sub Func_enable_cqweb_fts;
sub Func_disable_cqweb_fts;
sub Func_run_batch_idx;
sub Func_start_update_idx;
sub Func_update_fts_prop_files;
sub Func_stop_update_idx;
sub Func_start_fts_was_profile;
sub Func_fresh_batch_idx;
sub Func_fresh_update_idx;
sub Func_repair_records;
sub Func_optimize_idx;
sub Func_remove_lucene_idx_lock;
sub Func_add_record_type;
sub Func_create_fts_was_profile;
sub Func_delete_fts_was_profile;
sub Func_scrub_oplog;
sub Func_backup_fts;
sub Func_archive_fts;
sub Func_clear_state;
sub Func_lock_cq_fts;
sub Func_unlock_cq_fts;
sub Func_validate_cq_fts_deplo;
sub Func_prep_upgd_was_profiles;
sub Func_restore_was_profiles;
sub Func_gather_diagnostic_data;
sub Func_show_scenarios;
sub Func_set_was_max_mem;
sub Func_verbose;	# no-op

sub customizeFtsFiles;
sub run_installutil;
sub dieIfCqFtsIsLocked;
sub dieIfInState;
sub printValues;
sub setupCQFts;
sub doNothing;
sub getUserValue;
sub copyDataFile;
sub profileExist;
sub findNextPortNumberForProfile;
sub setDefaultValues;
sub runInstallUtilOnUnix;
sub runMultiutilOnUnix;
sub stopInstallUtilWithWindowsMsg;
sub removeRecordType;
sub createDeplySolrScrip;
sub deleteDeplySolrScrip;
sub createFTSConfigXMLFile;
sub isThisRecordTypeIndexed;
sub isRecordTypeExist;
sub isFieldListExist;

#
# deal with Unix / Windows
#
my $Q = "\'";	# assume Unix
my $gIsWindows = 0;
if ($^O =~ /MSWin/)
{
	$gIsWindows = 1;
	$Q = "\"";
}

my $ext = ($gIsWindows eq 1) ? ".bat" : ".sh";


#
# Create and set our global variables that are not config file driven
#
my $ftsClearQuestHomeDir = "";
$ftsClearQuestHomeDir = $ENV{'CQ_HOME'};
if (!defined($ftsClearQuestHomeDir) or $ftsClearQuestHomeDir eq "")
{
	$ftsClearQuestHomeDir = $ENV{'CLEARqUEST_HOME'};
}
if ((substr($ftsClearQuestHomeDir, -1) ne "/") or (substr($ftsClearQuestHomeDir, -1) ne "\\"))
{
	$ftsClearQuestHomeDir .= "/";
}
my $ftsRationalCommon = "";
$ftsRationalCommon = $ENV{'RATIONAL_COMMON'};
if (!defined($ftsRationalCommon) or $ftsRationalCommon eq "")
{
	$ftsRationalCommon = $ftsClearQuestHomeDir."../common";
}
if ((substr($ftsRationalCommon, -1) ne "/") or (substr($ftsRationalCommon, -1) ne "\\"))
{
	$ftsRationalCommon .= "/";
}
my $ftsJREHomeDir = "";
$ftsJREHomeDir = $ENV{'JRE_HOME'};
if ((!defined($ftsJREHomeDir) or $ftsJREHomeDir eq "") and $gIsWindows eq 0)
{
	$ftsJREHomeDir = $ftsClearQuestHomeDir."../common/java/jre/";
}
if ((substr($ftsJREHomeDir, -1) ne "/") or (substr($ftsJREHomeDir, -1) ne "\\"))
{
	$ftsJREHomeDir .= "/";
}
my $ftsCQtsdbcrawlerPL = $ftsClearQuestHomeDir."cqtsdbcrawler.pl";
my $ftsCqtssetupPL = $ftsClearQuestHomeDir."cqtssetup.pl";
my $ftsSetSolrHomePY = $ftsRationalCommon."eWAS/bin/setSolrHome.py";
my $ftsJavaEXE = $ftsJREHomeDir."bin/java";
my $ftsCQPerl = "";
my $ftsLockFile = "";
my $ftsStateFile = "";
my $ftsKeyFile = "";
my $ftsPWDFile = "";
my $ftsSourceDir = "";
my $ftsDBSetName = $optDbSet;
my $ftsUserDbName = $optuserDB;
my $ftsSiteName = "<local>";
my $ftsIsReplicated = 0;
my $ftsIsSiteWorkingMaster = 0;
my $ftsLocalReplicaName = $ftsSiteName;
my $ftsReplicaNames = "";
my $ftsReplicaNamesList = "";
my $ftsEnabledAtReplicaNamesList = "";
my $ftsUserName = $optUserName;
my $ftsUserPassword = $optPassword;
my $ftsTargetDir = "";
my $ftsWASProfileName;
my $ftsFtsWASProfilePort;
my $ftsFtsServerName = lc(hostname);
my $ftsUserDefinedEntityFile;
my $ftsDataDir = "";
my $ftsSolrFile = "";
my $ftsLuceneIndexDir = "";
my $ftsEntityFileName;
my $ftsEntityFile;
my $ftsCQPropertyFileName;
my $ftsCQPropertyFile;
my $ftsSolrIntermediateFileName;
my $ftsSolrIntermediateFile;
my $ftsIsEnableCQFTS;
my $ftsDoBatchIndex;
my $ftsStartupdateIndex;
my $ftsBatchIndexJVMParm;
my $ftsWasProfileMaxHeapSize;
my $ftsBatchIndexBatchSize;
my $ftsBatchIndexDelay;
my $ftsBatchIndexThreads;
my $ftsUpdateIndexBatchSize;
my $ftsUpdateIndexDelay;
my $ftsUpdateIndexLoginInterval;
my $ftsAddRecordType;
my $ftsRemoveRecordType;
my $ftsDeleteFtsWASProfiles;
my $ftsListOfIndxedRecordTypes;
my %ftsEntityFieldList = ();
my $installUtilMsgFile;
my $multiutilMsgFile;


fixPathSlashes();		# changes any "\" to "/"
validateUser(); 		# process only if user has proper CQ permission


#
# it's time to do real work, where data may get modified
#

# first, check to see if they are asking us to initialize, for the first time, a CQ FTS deployment
if ($optInitCqFTS eq "1")	# --init_cq_fts
{
	# we are asked to create a new depolyment
	initCqFTS();
	
	# with the --init_cq_fts option, they can also specify --create_fts_was_profile
	if ($optCreateCqFtsWasProfile eq "1")
	{
		push(@gCmdToExecute, ["Func_"."create_fts_was_profile"]);		# --create_fts_was_profile
	}
}
else
{
	# we are going after a pre-existing depolyment, read its values
	$optCqFtsHomeDir = findExistingCqFtsSetup();
}

readConfigData($optCqFtsHomeDir);	# read the XML config file into memory
setDefaultValues();					# fill in defaults if missing from the config file


#
# Now that we have read our config data, set our global variables
#
$ftsSourceDir = $ftsClearQuestHomeDir."CQ.Solr/SAMPL";
$ftsDBSetName = getUserValue("dbSet");
$ftsUserDbName = getUserValue("userDB");
$ftsSiteName = HTML::Entities::decode(getUserValue("siteName"));
$ftsEnabledAtReplicaNamesList = HTML::Entities::decode(getUserValue("enableAtSites"));
if ($ftsEnabledAtReplicaNamesList eq "")
{
	$ftsEnabledAtReplicaNamesList = $ftsReplicaNamesList;
}
else
{
	$ftsEnabledAtReplicaNamesList = cleanArrayList($ftsEnabledAtReplicaNamesList);
}
$ftsTargetDir = getUserValue("destination");
$ftsUserDefinedEntityFile = getUserValue("entityFile");
$ftsDataDir = resolveFTSDataDir();
$ftsLockFile = $ftsDataDir."cqfts.lock";
$ftsStateFile = $ftsDataDir."cqfts.state";
$ftsKeyFile = $ftsDataDir."key.txt";
$ftsPWDFile = $ftsDataDir."pwd.txt";
$ftsSolrFile = $ftsDataDir."Solr/solr/conf/schema.xml";
$ftsLuceneIndexDir = $ftsDataDir."Solr/solr/data/index/";
$ftsWASProfileName = getUserValue("ftsWASProfileName");
$ftsFtsWASProfilePort = getUserValue("ftsWASProfilePort");
$ftsFtsServerName = getUserValue("ftsServerName");
$ftsEntityFileName = "Entity-".$ftsDBSetName."-".$ftsUserDbName.".txt";
$ftsEntityFile = $ftsDataDir.$ftsEntityFileName;
$ftsCQPropertyFileName = "CQ-".$ftsDBSetName."-".$ftsUserDbName.".xml";
$ftsCQPropertyFile = $ftsDataDir.$ftsCQPropertyFileName;
$ftsSolrIntermediateFileName = "Solr-".$ftsDBSetName."-".$ftsUserDbName.".xml";
$ftsSolrIntermediateFile = $ftsDataDir.$ftsSolrIntermediateFileName;
$ftsIsEnableCQFTS = getUserValue("enableCQFTS");
$ftsDoBatchIndex = getUserValue("runBatchIndex");
$ftsStartupdateIndex = getUserValue("startUpdateIndex");
$ftsBatchIndexJVMParm = getUserValue("batchIndexJVMParm");
$ftsWasProfileMaxHeapSize = getUserValue("ftsWASProfileMaxHeapSize");
$ftsBatchIndexBatchSize = getUserValue("batchIndexBatchSize");
$ftsBatchIndexDelay = getUserValue("batchIndexDelay");
$ftsBatchIndexThreads = getUserValue("batchIndexThreads");
$ftsUpdateIndexBatchSize = getUserValue("updateIndexBatchSize");
$ftsUpdateIndexDelay = getUserValue("updateIndexDelay");
$ftsUpdateIndexLoginInterval = getUserValue("updateIndexLoginInterval");
$ftsAddRecordType = getUserValue("addRecordType");
$ftsRemoveRecordType = getUserValue("removeRecordType");
$ftsDeleteFtsWASProfiles = getUserValue("deleteFtsWASProfiles");

fixPathSlashes();			# changes any "\" to "/", do this again to clean up what we may have read from our config file
cleanupConfigUserData();	# the user typed data in the config file may need some cleanup
openLogger();				# we can now write to the log file

$installUtilMsgFile = $ftsDataDir."EnablingCQWebFTS.txt";
$multiutilMsgFile = $ftsDataDir."ScruboplogFTS.txt";

printAndLog $msgLabel_INFO."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CONFIG_XML_FILE", $optCqFtsHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_HOME", $ftsClearQuestHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_COMMON_HOME", $ftsRationalCommon)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JRE_HOME", $ftsJREHomeDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_USER_NAME", $ftsUserName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CQ_USER_PW", "********")."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_SOURCE_TEMPL_DIR", $ftsSourceDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_TARGET_DIR", $ftsTargetDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_FTS_DATA_DIR", $ftsDataDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_DB_SET_NAME", $ftsDBSetName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_USER_DB_NAME", $ftsUserDbName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SITE_NAME", $ftsSiteName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SITE_NAME_LIST", $ftsReplicaNamesList)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENABLE_AT_SITES", $ftsEnabledAtReplicaNamesList)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_NAME", $ftsWASProfileName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_PORT", $ftsFtsWASProfilePort)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROFILE_XMX_SETTING", $ftsWasProfileMaxHeapSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SERVER_NAME", $ftsFtsServerName)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_USER_ENTITY_FILE", $ftsUserDefinedEntityFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SOLR_SCHEMA_FILE", $ftsSolrFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_LUCENE_IDX_DIR", $ftsLuceneIndexDir)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENTITY_FILE", $ftsEntityFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PROPERTY_FILE", $ftsCQPropertyFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SOLR_INTERMEDIATE_FILE", $ftsSolrIntermediateFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_PW_KEY_FILE", $ftsKeyFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENCRYPTED_PW_FILE", $ftsPWDFile)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_CRAWLER_FILE", $ftsCQtsdbcrawlerPL)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SETUP_FILE", $ftsCqtssetupPL)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_SETSOLRHOME_FILE", $ftsSetSolrHomePY)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JAVA_EXE_FILE", $ftsJavaEXE)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ENABLE_CQ_FTS", $ftsIsEnableCQFTS)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_START_BATCH_IDX", $ftsDoBatchIndex)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_START_UPDATE_IDX", $ftsStartupdateIndex)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_JVM_BATCH_MODE", $ftsBatchIndexJVMParm)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_SIZE", $ftsBatchIndexBatchSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_DELAY", $ftsBatchIndexDelay)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_BATCH_THREADS", $ftsBatchIndexThreads)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_SIZE", $ftsUpdateIndexBatchSize)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_DELAY", $ftsUpdateIndexDelay)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_UPDATE_LOGIN_INTERVAL", $ftsUpdateIndexLoginInterval)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_ADD_RECORD", $ftsAddRecordType)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_REMOVE_RECORD", $ftsRemoveRecordType)."\n";
printAndLog $msgLabel_INFO.$langMsgText->maketext("XINFO_DELETE_PROFILES", $ftsDeleteFtsWASProfiles)."\n";


# the next logic depends on the following rules which was checked earlier.
# at this point, they either have used --init_cq_fts or --setup_cq_fts or 
# --create_fts_was_profile, which are the only 3 that we allow them to mix.
# since we already processed --init_cq_fts, lets see if they have --setup_cq_fts
# if so, process it, if not, and if --init_cq_fts was not given, then we process
# what ever they gave us on the command line

# if they are asking for a default end-to-end setup, do so now
if ($optSetupCqFTS eq "1")	# --setup_cq_fts
{
	# use default end-to-end setup
	setupCQFts();
}
# otherwise, if they have not used '--init_cq_fts', than it means they are passing their own command(s), so run it/them
elsif ($optInitCqFTS ne "1")	# --init_cq_fts
{
	# build the list of command-line options the user wanted to execute
	@gCmdToExecute = ();
	
	foreach my $item (@gCmdLineArgvList)
	{
		if (($item eq "username") or ($item eq "password") or
			($item eq "dbset") or ($item eq "userdb") or
			($item eq "ftshome") or ($item eq "init_cq_fts") or
			($item eq "setup_cq_fts"))
		{
			next;
		}
		else
		{
			my $foundAt = index($item, '=');	# if the command takes arg, remove the the arg part
			if ($foundAt ne -1)
			{
				$item = substr($item, 0, $foundAt);
			}
			push(@gCmdToExecute, [ "Func_".$item ]);
		}
	}
}


#
# we are now ready to execute each command in the order given
#
foreach my $function (@gCmdToExecute)
{
	my $argsStr = "";
	my @args = ();
	my $skipFuncName = 1;
	foreach my $arg (@$function)
	{
		if ($skipFuncName eq 1)
		{
			$skipFuncName = 0;
			next;
		}
		
		push(@args, $arg);		# build the function args
		
		if (length($argsStr) ne 0)
		{
			$argsStr .= ", ";
		}
		$argsStr .= $arg;
	}
	
	my $funcName = @$function[0];
	my %funcCaller = ( $funcName => \&$funcName );
	
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CALLING_TEXT").$funcName."(".$argsStr.")"."\n";
	
	$funcCaller{$funcName}->(@args);	# make the function call
}


#
# If the 'AboutThisFTS.txt' file doesn't exist, create it and tell the user about it
#
$gAboutThisFTSFile = $ftsDataDir."AboutThisFTS.txt";
if (!(-f $gAboutThisFTSFile))
{
	# if we just executed the 'archive_fts' command, than we can't create the 'about' file
	if ($optArchiveCqFTS ne "1")
	{
		createAboutThisFts($gAboutThisFTSFile);
	}
}


if ($ftsIsSiteWorkingMaster eq 0)	# give them a warning about working master site
{
	printAndLog $msgLabel_WARN."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L1")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L2")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("AT_WOKING_MASTER_L3")."\n", 1;
	printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN."\n", 1;
}


if ($optArchiveCqFTS ne "1")	# don't show this msg if we are archiving
{
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SEE_FILE", $gAboutThisFTSFile)."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FOR_IMPORTANT_NOTES")."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO."\n";
}

closeLogger();			# close the logger

exit 0;		# exit script, all is well




##############################################################################
# Close the logger
sub closeLogger
{
	if ($cqtsadminLogFile ne "")
	{
		# finish off our log
		printAndLog $msgLabel_INFO.$langMsgText->maketext("LOG_ENDED_ON", "".localtime())."\n";
		close logFileHandler;
	}
	elsif ($logDataBacklog ne "")
	{
		if ($optArchiveCqFTS eq "")
		{
			# we run into issues creating our log file, so attempt to flush the log data (in memory) to local file
			my $flushLogFile = $ftsDataDir."log-flush.log";
			if (!open(FH, ">".$flushLogFile))	# yes, it's OK to write to the file as utf-8, it's what we want
			{
				warn $msgLabel_WARN.""."\n", 1;
				warn $msgLabel_WARN."**************** ".$langMsgText->maketext("NON_FATAL_ERROR")." ****************"."\n", 1,
				warn $msgLabel_WARN.$langMsgText->maketext("CANNOT_CREATE_LOG_FILE", $flushLogFile)." ".$msgLable_oS_ERROR.$!, 1;
				warn $msgLabel_WARN.$langMsgText->maketext("LOST_LOG_MSG_LINE_1")."\n", 1;
				warn $msgLabel_WARN.$langMsgText->maketext("LOST_LOG_MSG_LINE_2")."\n", 1;
				warn $msgLabel_WARN.""."\n", 1;
			}
			else
			{
				print FH $logDataBacklog;
				close(FH);
				warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG", $flushLogFile)."\n", 1;
			}
		}
	}
	
	$cqtsadminLogFile = "";	# close the logger
}


##############################################################################
# Creates the log file and opens it.
sub openLogger
{
	my $logDir = $ftsDataDir."logs/";
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	$cqtsadminLogFile = $logDir."cqtsadmin_log-".$ftsDBSetName."_".$ftsUserDbName."-".$timeData.".log";
	
	my $tmp = "";
	
	if (!(-e $logDir))
	{
		$tmp = $msgLabel_INFO."mkpath -> [".$logDir."]"."\n";
		File::Path::mkpath $logDir;
	}
	
	if (!open(logFileHandler, ">:utf8", $cqtsadminLogFile))	
	{
		warn $msgLabel_WARN.""."\n", 1;
		warn $msgLabel_WARN."**************** ".$langMsgText->maketext("NON_FATAL_ERROR")." ****************"."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("CANNOT_CREATE_LOG_FILE", $cqtsadminLogFile)."\n", 1;
		warn $msgLabel_WARN.$msgLable_oS_ERROR.$!."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG_ATTEMPT_LINE_1", $ftsDataDir."log-flush.log")."\n", 1;
		warn $msgLabel_WARN.$langMsgText->maketext("FLUSH_LOG_ATTEMPT_LINE_2")."\n", 1;
		warn $msgLabel_WARN.""."\n", 1;
		
		$cqtsadminLogFile = "";
	}
	else
	{
		logFileHandler->autoflush(1);
	}
	
	printAndLog $tmp;
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOG_STARTED_ON", "".localtime())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("WRITTING_TO_LOG_FILE", $cqtsadminLogFile)."\n";
}


##############################################################################
# Prints a message tot he screen and logs it
sub printAndLog
{
	my ($msg, $sendToScreen) = @_;
	
	if (!defined($sendToScreen))
	{
		$sendToScreen = 0;	# if we were not told to display the text, we won't
	}
	
	if (($sendToScreen eq 0) and ($msg =~ m/====>/))
	{
		$sendToScreen = 1;	# always display banner text
	}
	
	if (($gVerboseOutput eq 1) or ($sendToScreen eq 1))
	{
		my $screenMsg = $msg;
		
		if (!defined($::PRINTANDLOG_ENC) and defined($gCqFtsEncoding))
		{
			if ($gCqFtsEncoding ne "")
			{
				$screenMsg = Encode::encode($gCqFtsEncoding, $screenMsg);
			}
		}

		print $screenMsg;		# output the message to the screen.
	}
	
	if (defined($::PRINTANDLOG_ENC) and defined($gCqFtsEncoding))
	{
		if ($gCqFtsEncoding ne "")
		{
			$msg = Encode::decode($gCqFtsEncoding, $msg);
		}
	}
	
	if ($cqtsadminLogFile eq "")
	{
		# the log file is not ready yet, cache log data into memory
		$logDataBacklog .= $msg;
	}
	else
	{
		# the log file is ready, write to it
		if ($logDataBacklog ne "")
		{
			# but first flush what we haev in memory
			print logFileHandler $logDataBacklog;
			$logDataBacklog = "";
		}
		
		print logFileHandler $msg;
	}
}


##############################################################################
# Reports an error message and terminates the script (if told so)
sub dealWithError
{
	my ($errMsg, $dieOnError) = @_;
	
	if ($dieOnError)
	{
		printAndLog "\n".$gTerminationErrorMsg."\n", 1;		# always show error messages, independet of if "verbose" is set or not
		printAndLog "\n".$errMsg, 1;
		
		die "\n"; ### ### .$errMsg;
	}
	else
	{
		printAndLog $msgLabel_WARN."*********vvvvvv ".$langMsgText->maketext("IGNORE_SAFE_ERROR")." vvvvvv*********"."\n", 1;
		printAndLog $errMsg, 1;
		printAndLog $msgLabel_WARN."*********^^^^^^ ".$langMsgText->maketext("IGNORE_SAFE_ERROR")." ^^^^^^*********"."\n", 1;
	}
}


##############################################################################
# Changes the current working directory and verifies the change
sub chdirAndValidate
{
	my ($chdirTo, $dieOnError) = @_;
	
	printAndLog $msgLabel_INFO."chdir -> [".$chdirTo."]"."\n";
	
	my $status = chdir $chdirTo;
	if (! $status)
	{
		dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  chdir ".$chdirTo."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Checks to see if a file was created or not
sub validateFileGenerated
{
	my ($executedCmd, $expectedFile, $dieOnError) = @_;
	
	# mask out password string
	my $msg = $executedCmd;
	
	$msg =~ s/-p \Q$ftsUserPassword\E/-p ********/sg;
	$msg =~ s/$ftsUserName \Q$ftsUserPassword\E/$ftsUserName ********/sg;
	
	my $status = 1;
	
	if (!(-f $expectedFile))
	{
		my $errMsg = $langMsgText->maketext("ERROR_DURING_EXE")."\n"."  ".$msg."\n\n".$langMsgText->maketext("FILE_NOT_GENERATED")."\n";
		
		dealWithError($errMsg, $dieOnError);
		
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Executes an external program / command; checks for it's success / completion, 
# and terminates if it fails (if instructed to)
sub executeExternalCommand
{
	my ($cmd, $matchingString, $dieOnErrorMatchKind, $dieOnError) = @_;
	
	# mask out password string from the command to be executed
	my $maskedCmd = $cmd;
	
	$maskedCmd =~ s/-p \Q$ftsUserPassword\E/-p ********/sg;
	$maskedCmd =~ s/$ftsUserName \Q$ftsUserPassword\E/$ftsUserName ********/sg;
	
	printAndLog $msgLabel_INFO."system -> [".$maskedCmd."]"."\n";
	
	my $status = 1;
	
	# execute the command
	if (open(FH,  "$cmd 2>&1 |"))
	{
		my $foundMatchingString = 0;
		my $line;
		
		while ($line = <FH>)
		{
			# output to the screen and log file unless if they asked not to (undocumented)
			if (not (defined($gSuppressExternalCmdOutput) and $gSuppressExternalCmdOutput eq "yes"))
			{
				if (defined($gExternalCmdMarker) and $gExternalCmdMarker ne "")
				{
					printAndLog $gExternalCmdMarker;
				}
				
				{
					local $::PRINTANDLOG_ENC = $gCqFtsEncoding;
					printAndLog $line;
				}
			}
			
			foreach my $matchItem (@$matchingString)
			{
				if (($matchItem ne "") and ($line =~ /$matchItem/))
				{
					$foundMatchingString = 1;
				}
			}
		}
		close FH;
		
		if ($matchingString ne "")
		{
			if (($foundMatchingString eq 1) and ($dieOnErrorMatchKind eq $gDieKind_must_not_match))
			{
				# we found the required string, but we are not suppose to, so we will die
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
			elsif (($foundMatchingString eq 1) and ($dieOnErrorMatchKind eq $gDieKind_must_match))
			{
				# we found the required string, which we are required to, so we will not die
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_must_not_match))
			{
				# we did not find our matching string, which we are not suppose to, so we don't die
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_must_match))
			{
				# we did not find our matching string, which we are suppose to, so we will die
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
			elsif (($foundMatchingString eq 0) and ($dieOnErrorMatchKind eq $gDieKind_none))
			{
				# we did not find the required string, which we are not required to, so we don't die
			}
			else
			{
				# this is an unknown state ... 
				dealWithError($langMsgText->maketext("ERROR_DURING_EXE")."\n"."  CMD: [".$maskedCmd."]"."\n", $dieOnError);
				$status = 0;
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_EXE", $cmd)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Creates the file, "AboutThisFTS.txt' file
sub createAboutThisFts
{
	my ($aboutThisFtsFile) = @_;
	
	if (!open(FH, ">:utf8", $aboutThisFtsFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $aboutThisFtsFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	print FH $langMsgText->maketext("ABOUT_INFO")."\n";
	print FH "\n";
	print FH $langMsgText->maketext("ABOUT_CREATED_ON", "".localtime())."\n";
	print FH $langMsgText->maketext("ABOUT_OS", $^O)."\n";
	print FH $langMsgText->maketext("ABOUT_CFG_XML_FILE", $optCqFtsHomeDir)."\n";
	print FH $langMsgText->maketext("ABOUT_SOURCE_TEMPL_DIR", $ftsSourceDir)."\n";
	print FH $langMsgText->maketext("ABOUT_DB_SET_NAME", $ftsDBSetName)."\n";
	print FH $langMsgText->maketext("ABOUT_USER_DB_NAME", $ftsUserDbName)."\n";
	print FH $langMsgText->maketext("ABOUT_SITE_NAME", $ftsSiteName)."\n";
	print FH $langMsgText->maketext("ABOUT_SITE_LIST_NAME", $ftsReplicaNamesList)."\n";
	print FH $langMsgText->maketext("ABOUT_ENABLED_SITE_LIST_NAME", $ftsEnabledAtReplicaNamesList)."\n";
	print FH $langMsgText->maketext("ABOUT_SERVER_HOST", $ftsFtsServerName)."\n";
	print FH $langMsgText->maketext("ABOUT_FTS_TARGET_DIR", $ftsTargetDir)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_NAME", $ftsWASProfileName)."\n";
	print FH $langMsgText->maketext("ABOUT_PROFILE_PORT", $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_ADMIN_URL", $ftsFtsServerName, $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_LUCENE_INDEX_LOCATION", $ftsLuceneIndexDir)."\n";
	print FH $langMsgText->maketext("ABOUT_DATA_DIR", $ftsDataDir)."\n";
	print FH $langMsgText->maketext("ABOUT_ENTITY_FILE", $ftsEntityFile)."\n";
	print FH $langMsgText->maketext("ABOUT_PROPERTY_FILE", $ftsCQPropertyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_INTERM_FILE", $ftsSolrIntermediateFile)."\n";
	print FH $langMsgText->maketext("ABOUT_PW_KEY_FILE", $ftsKeyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_ENCRPTED_PW_FILE", $ftsPWDFile)."\n";
	print FH $langMsgText->maketext("ABOUT_SOLR_FILE", $ftsSolrFile)."\n";
	print FH "\n"."\n";
	print FH $langMsgText->maketext("ABOUT_POST_SETUP_INFO")."\n";
	print FH "\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_1", $ftsFtsServerName, $ftsFtsWASProfilePort)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_2", $ftsCQPropertyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_3", $ftsKeyFile)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_4", $ftsWASProfileName)."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_5")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_6")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_7")."\n";
	print FH $langMsgText->maketext("ABOUT_POST_INFO_8")."\n";
	
	if ($gIsWindows eq 0)
	{
		print FH $langMsgText->maketext("ABOUT_POST_INFO_9", $installUtilMsgFile)."\n";
		runInstallUtilOnUnix();
		print FH $langMsgText->maketext("ABOUT_POST_INFO_10", $multiutilMsgFile)."\n";
		runMultiutilOnUnix();
	}
	
	print FH "\n"."\n";
	print FH $langMsgText->maketext("ABOUT_SEE_INFO_CENTER")."\n";
	
	close FH;
}


##############################################################################
# Fixes up the FTS home path
sub resolveFTSDataDir
{
	if ($ftsDataDir eq "/")
	{
		my $foundAt = rindex($optCqFtsHomeDir, "/") + 1;
		if ($foundAt eq -1)
		{
			$foundAt = rindex($optCqFtsHomeDir, "\\") + 1;
		}
		if ($foundAt ne -1)
		{
			$ftsDataDir = substr($optCqFtsHomeDir, 0, $foundAt);
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_DETERMINE_FTS_DIR", $optCqFtsHomeDir)."\n", 1);
		}
	}
	
	return ($ftsDataDir);
}


##############################################################################
# Creates a new FTS setup
sub createNewCqFtsSetup
{
	if (substr($optCqFtsHomeDir, -1) ne "/")
	{
		$optCqFtsHomeDir .= "/";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATE_NEW_CQFTS", $optCqFtsHomeDir)."\n";
	
	$ftsTargetDir = $optCqFtsHomeDir;	# this is now our FTS target setup directory
	
	$ftsDataDir = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/";
	printAndLog $msgLabel_INFO."mkpath -> [".$ftsDataDir."]"."\n";
	File::Path::mkpath $ftsDataDir;
	
	$optCqFtsHomeDir = $ftsDataDir."cqtsadmin"."-".$ftsDBSetName."-".$ftsUserDbName.".xml";
	createFTSConfigXMLFile($optCqFtsHomeDir);
}


##############################################################################
# Looks to see if an existing FTS setup exist or not
sub findExistingCqFtsSetup
{
	my $result = "";
	
	$ftsTargetDir = $optCqFtsHomeDir;	# this is now our FTS target setup directory
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOCATE_CQFTS", $ftsTargetDir)."\n";
	
	# folder exist, look for the default file CQ FTS XML configuration file
	$ftsDataDir = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/";
	$result = $ftsDataDir."cqtsadmin"."-".$ftsDBSetName."-".$ftsUserDbName.".xml";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_FOR_CQFTS", $result)."\n";
	if (!(-f $result))		# "/CQ.Search/TextSearch_SAMPL/cqtsadmin-TextSearch-SAMPL.xml"
	{
		$result = "";
	}
	
	return ($result);
}


##############################################################################
# Changes '\' to '/' and adds a '/' to the end of the path (if one doesn't exist)
sub fixPathSlashes
{
	$ftsClearQuestHomeDir =~ s/\\/\//sg;
	$ftsRationalCommon =~ s/\\/\//sg;
	$ftsJREHomeDir =~ s/\\/\//sg;
	$ftsSourceDir =~ s/\\/\//sg;
	$ftsTargetDir =~ s/\\/\//sg;
	$ftsDataDir =~ s/\\/\//sg;
	$ftsCqtssetupPL =~ s/\\/\//sg;
	$ftsCQtsdbcrawlerPL =~ s/\\/\//sg;
	$ftsSetSolrHomePY =~ s/\\/\//sg;
	$ftsSolrFile =~ s/\\/\//sg;
	$ftsLuceneIndexDir =~ s/\\/\//sg;
	$ftsCQPerl =~ s/\\/\//sg;
	$ftsJavaEXE =~ s/\\/\//sg;
	$gPrepUpgdFtsWASProfiles =~ s/\\/\//sg;
	$gRestoreFtsWASProfiles =~ s/\\/\//sg;
	
	if (substr($ftsClearQuestHomeDir, -1) ne "/")
	{
		$ftsClearQuestHomeDir .= "/";
	}
	if (substr($ftsRationalCommon, -1) ne "/")
	{
		$ftsRationalCommon .= "/";
	}
	if (substr($ftsJREHomeDir, -1) ne "/")
	{
		$ftsJREHomeDir .= "/";
	}
	if (substr($ftsSourceDir, -1) ne "/")
	{
		$ftsSourceDir .= "/";
	}
	if (substr($ftsTargetDir, -1) ne "/")
	{
		$ftsTargetDir .= "/";
	}
	if (substr($ftsDataDir, -1) ne "/")
	{
		$ftsDataDir .= "/";
	}
	if (substr($gPrepUpgdFtsWASProfiles, -1) ne "/")
	{
		$gPrepUpgdFtsWASProfiles .= "/";
	}
	if (substr($gRestoreFtsWASProfiles, -1) ne "/")
	{
		$gRestoreFtsWASProfiles .= "/";
	}
}


##############################################################################
# After reading cqfts admin data, clean it up before usage
sub cleanupConfigUserData
{
	# if they didn't give us any passowrd, it means no password which is a blank quote
	if ($ftsUserPassword eq "")
	{
		$ftsUserPassword = "\"\"";
	}
	
	# see if they want to associate the db-set and user-db name with the profile name
	$ftsWASProfileName =~ s/\[user-db-name\]/$ftsUserDbName/s;
	$ftsWASProfileName =~ s/\[db-set-name\]/$ftsDBSetName/s;
	
	# work out the port for the WAS profile
	if ($ftsFtsWASProfilePort eq "[*]")
	{
		# the CQ FTS config XML file doesn't have a port number, figure it out
		if ($optCreateCqFtsWasProfile eq "1")
		{
			# we are createing a new profile, and are asked to figure out it's port number
			$ftsFtsWASProfilePort = findNextPortNumberForProfile();
		}
		else
		{
			# we are asked to get the port number of a given profile
			$ftsFtsWASProfilePort = findPortNumberForProfile($ftsWASProfileName);
			if ($ftsFtsWASProfilePort eq "-1")
			{
				printAndLog "\n".$gTerminationErrorMsg."\n", 1;
				dealWithError($langMsgText->maketext("RETRIEVE_WAS_PORT_NUMBER", $ftsWASProfileName)."\n", 1);
			}
		}
	}
	
	if (length($ftsWASProfileName) > $gMaxWASProfileLen)
	{
		printAndLog "\n".$gTerminationErrorMsg."\n", 1;
		dealWithError($langMsgText->maketext("PROFILE_NAME_TOO_LONG", $ftsWASProfileName, $gMaxWASProfileLen)."\n", 1);
	}
	
	# remove any spaces that maybe in the list of record-types to add / remove
	$ftsAddRecordType = trim($ftsAddRecordType);
	$ftsRemoveRecordType = trim($ftsRemoveRecordType);
	$ftsDeleteFtsWASProfiles = trim($ftsDeleteFtsWASProfiles);
}


##############################################################################
# Verify that the CQ user has proper permission
sub validateUser
{
	my $user = $ftsUserName;
	my $pwd = $ftsUserPassword;
	my $dbSet = $ftsDBSetName;
	my $dbName = $ftsUserDbName;
	
	if ($pwd eq "\"\"")
	{
		$pwd = "";
	}
	
	eval
	{
		my $adminSessionObj = CQAdminSession::Build();
		
		$adminSessionObj->Logon($user, $pwd, $dbSet);
		
		$ftsIsReplicated = $adminSessionObj->IsReplicated();
		$ftsReplicaNames = $adminSessionObj->GetReplicaNames();
		$ftsLocalReplicaName = $adminSessionObj->GetLocalReplicaName();
		$ftsIsSiteWorkingMaster = $adminSessionObj->IsSiteWorkingMaster();
		
		CQAdminSession::Unbuild($adminSessionObj);
	};
	die "\n"."$@\n" if $@; $@ = "";
	
	if ($ftsIsReplicated eq 1)
	{
		foreach my $replicName (@$ftsReplicaNames)
		{
			if ($ftsReplicaNamesList ne "")
			{
				$ftsReplicaNamesList .= ",";
			}
			$ftsReplicaNamesList .= trim($replicName);
		}
	}
	else
	{
		$ftsReplicaNamesList = "";
	}
	
	$ftsSiteName = $ftsReplicaNamesList;
	
	# User session, for doing actual work
	my $sessionObj;
	
	eval
	{
		$sessionObj = CQSession::Build();
	
		$sessionObj->UserLogon($user, $pwd, $dbName, $dbSet);
	};
	die "\n"."$@\n" if $@; $@ = "";
	
	if ($sessionObj->IsUserSuperUser() eq 0)
	{
		dealWithError($langMsgText->maketext("NO_PRIVILAGE")."\n", 1);
	}
	
	my $productInfo = $sessionObj->GetProductInfo();	
	$cqVersion = $productInfo->GetFullProductVersion();
	
	my $indexedRecordTypes = "";
	
	if (!($cqVersion lt 7.1))
	{
		$isFTSEnabled = $sessionObj->IsTextSearchEnabled();
		if ($isFTSEnabled)
		{
			my $textSearchDef = $sessionObj->BuildTextSearch();
			$ftsListOfIndxedRecordTypes = $textSearchDef->GetIndexedEntityDefNames();
			
			my $recordType;
			my $indexedRecordTypes = "";
			foreach $recordType (@$ftsListOfIndxedRecordTypes)
			{
				if ($indexedRecordTypes ne "")
				{
					$indexedRecordTypes .= ", ";
				}
				$indexedRecordTypes .= $recordType;
			}
		}
		else
		{
			$indexedRecordTypes = $langMsgText->maketext("FTS_NOT_ENABLED");
			$ftsListOfIndxedRecordTypes = "";
		}
	}
	
	# check to see if CQ FTS is installed
	if (!(-f $ftsCqtssetupPL))
	{
		dealWithError($langMsgText->maketext("CQ_FTS_NOT_INSTALLED")."\n", 1);
	}
	
	{
		my $entityDefNames = $sessionObj->GetSubmitEntityDefNames();
		
		# Iterate over the record types
		foreach my $entityName ( @$entityDefNames )
		{
			my $entDef = $sessionObj->GetEntityDef($entityName);
			
			my $fieldsListString = "";
			my ($fieldNamesRef) = $entDef->GetFieldDefNames();
			foreach my $fieldName (@$fieldNamesRef)
			{
				if ($fieldName eq "history")
				{
					next;	# 'history' field is not FTS indexed
				}
				
				$fieldsListString = $fieldsListString.",".$fieldName;
			}
			
			$ftsEntityFieldList{$entDef->GetName()} = $fieldsListString;
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FTS_INDEXED_RECORDS_TYPES", $indexedRecordTypes)."\n";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("CQ_PROD_INFO")."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FULL_PROD_VER", $productInfo->GetFullProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("PROD_VER", $productInfo->GetProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("PATCH_VER", $productInfo->GetPatchVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LICE_VER", $productInfo->GetLicenseVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LICE_FEATURE", $productInfo->GetLicenseFeature())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("WEB_LICE_VER", $productInfo->GetWebLicenseVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_NAME", $productInfo->GetCompanyName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_FULL_NAME", $productInfo->GetCompanyFullName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_EMAIL_ADDESS", $productInfo->GetCompanyEmailAddress())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("COMPANY_WEB_ADDRESS", $productInfo->GetCompanyWebAddress())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BUILD_NUM", $productInfo->GetBuildNumber())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("STAGE_LABEL", $productInfo->GetStageLabel())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DEFAULT_DBSET_NAME", $productInfo->GetDefaultDbSetName())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("OM_VER_MAJOR", $productInfo->GetObjectModelVersionMajor())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("OM_VER_MINOR", $productInfo->GetObjectModelVersionMinor())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SUITE_PROD_VER", $productInfo->GetSuiteProductVersion())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SUITE_BUILD_ID", $productInfo->GetSuiteBuildId())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FEATURE_LEVEL", $sessionObj->GetSessionFeatureLevel())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("IS_REPLICATED", $sessionObj->IsReplicated())."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("IS_FTS_ENABLED", $sessionObj->IsTextSearchEnabled())."\n";
	
	my $FL = $sessionObj->GetSessionFeatureLevel();
	my $gIsReplicated = $sessionObj->IsReplicated();
	if (($FL < 7) and ($gIsReplicated eq 0))
	{
		$gIsReplicated = ($gIsReplicated eq 1) ? $langMsgText->maketext("IS_MULTISITED") : $langMsgText->maketext("IS_NOT_MULTISITED");
		
		printAndLog $msgLabel_WARN.""."\n", 1;
		printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
		printAndLog $msgLabel_WARN.$langMsgText->maketext("DB_FL", $FL)."\n", 1;
		printAndLog $msgLabel_WARN.$gIsReplicated."\n", 1;
		printAndLog $msgLabel_WARN.$langMsgText->maketext("NEED_FL7_OR_MULTISITE")."\n", 1;
		printAndLog $msgLabel_WARN."************************************************************"."\n", 1;
		printAndLog $msgLabel_WARN.""."\n", 1;
	}
	
	CQSession::Unbuild($sessionObj);
	
	if ($cqVersion lt 7.1)	# make sure the tool is being run on the correct version of CQ
	{
		dealWithError($langMsgText->maketext("WRONG_CQ_VERSION", $cqVersion)."\n", 1);
	}
}


##############################################################################
# Helper function to recursively copy files
sub copyDataFile
{
	# make the data directory if it doesn't exist
	if (!(-e $ftsDataDir))
	{
		printAndLog $msgLabel_INFO."mkpath -> [".$ftsDataDir."]"."\n";
		File::Path::mkpath $ftsDataDir;
	}
	
	# copy the data file over
	my @fileName = split(/\//, $optCqFtsHomeDir);
	my $to = $ftsDataDir.$fileName[@fileName - 1];
	if (!(-e $to))
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CONFIG_IS_HERE", $to)."\n";
		if (!copy($optCqFtsHomeDir, $to))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $optCqFtsHomeDir, $msgLable_oS_ERROR)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Checks to see if a WAS profile exist or not
sub profileExist
{
	my ($theProfileName) = @_;
	
	my $wasProfiles = $ftsRationalCommon."CM/profiles/";
	my $doesProfileExist = 0;
	
	if (index($wasProfiles, ' ') ne -1)
	{
		$wasProfiles = $Q.$wasProfiles.$Q;
	}
	
	my $profileName;
	my @files = <$wasProfiles*>;
	foreach $profileName (@files)
	{
		my $currentProfileName = "";
		my $foundAt = rindex($profileName, "/") + 1;
		if ($foundAt eq -1)
		{
			$foundAt = rindex($profileName, "\\") + 1;
		}
		if ($foundAt ne -1)
		{
			$currentProfileName = substr($profileName, $foundAt);
			if ($currentProfileName eq $theProfileName)
			{
				$doesProfileExist = 1;
				last;
			}
		}
	}
	
	return $doesProfileExist;
}


##############################################################################
# Returns a list of existing WAS profiles
sub getListOfProfiles
{
	my $wasProfiles = $ftsRationalCommon."CM/profiles/";
	
	if (index($wasProfiles, ' ') ne -1)
	{
		$wasProfiles = $Q.$wasProfiles.$Q;
	}
	
	my $transportPortMatch = $langMsgText->maketext("HTTP_TRANSPORT_PORT")->[0];
	my @profilesList = ();
	my $profilePath;

	my @files = <$wasProfiles*>;
	foreach $profilePath (@files)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_IN", $profilePath)."\n";
		
		$profilePath =~ s/\\/\//sg;
		
		my $foundAt = rindex($profilePath, "/");
		if ($foundAt ne -1)
		{
			$foundAt += 1;
			
			my $profileName = substr($profilePath, $foundAt);
			my $webXMLFolder = findWebXMLFolder($profileName);
			
			if (-e $webXMLFolder)
			{
				push(@profilesList, $profilePath);
			}
		}
	}
	
	return @profilesList;
}


##############################################################################
# Returns a list of existing profiles directory
sub getListOfProfilesDir
{
	my ($rootDir) = @_;
	
	my @profilesDirList = ();
	my $DH;
	my $dir;
	
	unless (opendir $DH, $rootDir)
	{
		dealWithError($langMsgText->maketext("NO_SUCH_DIR_FILE", $rootDir)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	while ($dir = readdir $DH)
	{
		next if (($dir eq ".") or ($dir eq "..") or (-f $rootDir.$dir));
		push(@profilesDirList, $dir);
	}
	
	return @profilesDirList;
}


##############################################################################
# Find the next WAS prfile port number
sub findNextPortNumberForProfile
{
	my $wasProfiles = $ftsRationalCommon."CM/profiles/";
	
	if (index($wasProfiles, ' ') ne -1)
	{
		$wasProfiles = $Q.$wasProfiles.$Q;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("NEXT_AVAILABLE_PORT")."\n";
	
	my $transportPortMatch = $langMsgText->maketext("HTTP_TRANSPORT_PORT")->[0];
	my $portNumber = "-1";
	my $profileName;
	my @files = <$wasProfiles*>;
	foreach $profileName (@files)
	{
		my $aboutFile = $profileName."/logs/AboutThisProfile.txt";
		my $line;
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_IN", $aboutFile)."\n";
		
		if (!open(FH, $aboutFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $aboutFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			if ($gCqFtsEncoding ne "")
			{
				$line = Encode::decode($gCqFtsEncoding, $line);
			}
			
			if ($line =~ /$transportPortMatch/)
			{
				my $foundAt = rindex($line, " ");
				if ($foundAt ne -1)
				{
					$line = substr($line, $foundAt);
					
					my $newPortNumber = trim($line);
					if ($newPortNumber gt $portNumber)
					{
						$portNumber = $newPortNumber;
					}
				}
			}
		}
		close FH;
	}
	
	$portNumber += 1000;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FOUND_NEXT_PORT", $portNumber)."\n";
	
	return $portNumber;
}


##############################################################################
# Returns the port number of a WAS profile
sub findPortNumberForProfile
{
    my ($theProfileName) = @_;
	
	my $wasProfiles = $ftsRationalCommon."CM/profiles/";
	
	if (index($wasProfiles, ' ') ne -1)
	{
		$wasProfiles = $Q.$wasProfiles.$Q;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_FOR_PORT", $theProfileName)."\n";
	
	my $transportPortMatch = $langMsgText->maketext("HTTP_TRANSPORT_PORT")->[0];
	my $portNumber = "-1";
	my $profileName;
	my @files = <$wasProfiles*>;
	foreach $profileName (@files)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("LOOKING_IN", $profileName)."\n";
	
		if ($profileName =~ /$theProfileName/)
		{
			my $aboutFile = $profileName."/logs/AboutThisProfile.txt";
			my $line;
			
			if (!open(FH, $aboutFile))
			{
				dealWithError($langMsgText->maketext("CANNOT_OPEN", $aboutFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
			}
			while ($line = <FH>)
			{
				if ($gCqFtsEncoding ne "")
				{
					$line = Encode::decode($gCqFtsEncoding, $line);
				}
				
				if ($line =~ /$transportPortMatch/)
				{
					my $foundAt = rindex($line, " ");
					if ($foundAt ne -1)
					{
						$line = substr($line, $foundAt);
						
						$portNumber = trim($line);
						last;
					}
				}
			}
			close FH;
		}
	}
	
	return $portNumber;
}

##############################################################################
# Copies files that match a given extension
sub copyFilesMatch
{
	my ($fromDir, $toDir, $extension, $dieOnError) = @_;
	
	my $status = 1;
	
	my @files; 
	if (opendir(DIR, $fromDir))
	{
		while (my $fileName = readdir(DIR))
		{
			if ($fileName =~ m/.$extension$/)
			{
				my $fromFile;
				my $toFile;
				
				$fromFile = $fromDir."/".$fileName;
				$toFile = $toDir."/".$fileName;
				
				if (-f $fromFile)
				{
					if (!copy($fromFile, $toFile))		# returns 0 on failure
					{
						dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromFile, $toFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
						$status = 0;
					}
				}
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $fromDir)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}

##############################################################################
# Recursively, copies files and sub-directories from source to destination
sub copyFiles
{
	my ($fromDir, $toDir, $dieOnError) = @_;
	
	my $status = 1;
	
	printAndLog $msgLabel_INFO."mkpath -> [".$toDir."]"."\n";
	File::Path::mkpath $toDir;
	
	if (opendir(DIR, $fromDir))
	{
		my @files = readdir DIR;
		closedir (DIR);	
		
		my $fromFile;
		my $toFile;
		my $file;
		
		foreach $file (@files)
		{
			$fromFile = $fromDir.$file;
			$toFile = $toDir.$file;
			if (-f $fromFile)
			{
				if (!copy($fromFile, $toFile))		# returns 0 on failure
				{
					dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromFile, $toFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
					$status = 0;
				}
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $fromDir)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Normalizes the Entity file so that any white spaces are eliminated
sub cleanupEntityFile
{
	my ($entityFile) = @_;
	
	my @data;
	if (!open(FH, $entityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $entityFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	@data = <FH>;
	close(FH);
	
	my $resultData;
	
	my $line;
	foreach $line (@data)
	{
		my @list = split(/=/, $line);
		
		if (scalar(@list) ne 2)
		{
			next;
		}
		
		my $entity = trim($list[0]);
		
		if ($entity eq "")
		{
			next;
		}
		
		my $fields = cleanupFieldsList($list[1]);
		
		$fields =~ s/,+/,/g;	# remove any extra double "," (blank field-names)
		$fields =~ s/^,//g;		# remove any extra start of ','
		$fields =~ s/,$//g;		# remove any extra ending of ','
		
		$resultData .= $entity."=".$fields."\n";
	}
	
	# Now write the changes to the file
	return safeFileUpdate($entityFile, $resultData, 0);
}


##############################################################################
# If the field-list of the Entity file has '&', clean it
sub cleanupFieldsList
{
	my ($fieldsList) = @_;
	
	my $cleanFieldsList = "";
	
	my @fields = split(/,/, $fieldsList);
	
	foreach my $field (@fields)
	{
		my $cleanField = trim($field);
		
		if (($cleanField eq "") or ($cleanField eq "&"))
		{
			next;
		}
		
		if (($field eq "locked_by") and (($cqVersion eq "7.1") or ($cqVersion eq "7.1.0") or ($cqVersion eq "7.1.0.0")))
		{
			# do not add 'locked_by' field: workaround known defect: RATLC01276003
			next;
		}
		
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		
		$cleanField =~ s/^&\s+/&/;	# remove any space that maybe between "&" and the field name
		
		$cleanFieldsList .= $cleanField;
	}
	
	# if missing, add required fields: dbid, record_type, version
	if (index($cleanFieldsList, "dbid") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "dbid";
	}
	if (index($cleanFieldsList, "record_type") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "record_type";
	}
	if (index($cleanFieldsList, "version") eq -1)
	{
		if ($cleanFieldsList ne "")
		{
			$cleanFieldsList .= ",";
		}
		$cleanFieldsList .= "version";
	}
	
	return $cleanFieldsList;
}


##############################################################################
# Determines if the list of fields exist (in the CQ DB) or not
sub isFieldListExist
{
	my ($recordTypeName, $fieldsList, $dieOnError) = @_;
	
	my @userGivenfieldsList = split(/,/, $fieldsList);
	
	my $found = 0;
	
	my $item;
	foreach  $item (keys %ftsEntityFieldList) 
	{
		my $entityName = $item;
		my $fieldList = $ftsEntityFieldList{$item};
		
		if ($entityName eq "history")
		{
			next;	# 'history' record-type is not indexed
		}
		
		if ($recordTypeName eq $entityName)
		{
			my @dbfieldsList = split(/,/, $fieldList);
			
			my $rt1;
			foreach $rt1 (@userGivenfieldsList)	# walk over the user list
			{
				$rt1 =~ s/^&//;		# remove "&" from the start filed name (if any)
				
				my $rt2;
				foreach $rt2 (@dbfieldsList)		# make sure it's in the DB's list
				{
					if ($rt1 eq $rt2)
					{
						$found++;
						last;
					}
				}
			}
			
			if ($found ne scalar @userGivenfieldsList)
			{
				$found = 0;
			}
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FIELD_NAME_NOT_IN_SCHEMA", $recordTypeName)."\n", 1);
	}
	
	return $found;
}


##############################################################################
# Determines if the list of record-types exist (in the CQ DB) or not
sub isRecordTypeExist
{
	my ($recordTypeName, $dieOnError) = @_;
	
	my $found = 0;
	
	my $item;
	foreach  $item (keys %ftsEntityFieldList) 
	{
		my $entityName = $item;
		
		if ($entityName eq "history")
		{
			next;	# 'history' field is not indexed
		}
		
		if ($recordTypeName eq $entityName)
		{
			$found = 1;
			last;
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("RECORD_TYPE_NOT_IN_SCHEMA", $recordTypeName)."\n", 1);
	}
	
	return $found;
}


##############################################################################
# Determines if given record-type is indexed or not
sub isThisRecordTypeIndexed
{
	my ($lookingFor, $dieOnError) = @_;
	
	my $found = 0;
	my $recordType;
	my $indexedRecordTypes = "";
	foreach $recordType (@$ftsListOfIndxedRecordTypes)
	{
		if ($recordType eq $lookingFor)
		{
			$found = 1;
			last;
		}
	}
	
	if (($found eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("RECORD_TYPE_NOT_INDEXED", $lookingFor)."\n", 1);
	}
	
	return $found;
}


##############################################################################
# Updates a file in a safe way; it will make a backup of the current file, 
# write to it, and delete the backup file.
sub safeFileUpdate
{
	my ($fileToUpdate, $data, $dieOnError) = @_;
	
	my $backupFile = $fileToUpdate."backup";
	my $status = 0;		# assume we will fail
	
	# first make sure the file exist
	if (!(-f $fileToUpdate))
	{
		dealWithError($langMsgText->maketext("UPDATE_FILE_NOT_EXIST", $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n", 0);
	}
	else
	{
		# make a backup copy of the current file (the one we will be modifying)
		if (!copy($fileToUpdate, $backupFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $fileToUpdate, $backupFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
		else
		{
			# now open the existing file and write to it the new data
			if (open(FH, ">$fileToUpdate"))
			{
				print FH $data;		# write to it and close it
				close FH;
				
				$status = 1;	# all is well
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_WRITE_TO_FILE", $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				
				# restore from the backup file
				if (!copy($backupFile, $fileToUpdate))
				{
					my $msg = $langMsgText->maketext("CANNOT_RESTORE_BACKUP_FILE", $backupFile, $fileToUpdate)."\n".$msgLable_oS_ERROR.$!."\n";
					$msg .= "\n\n".$langMsgText->maketext("MANUALY_RESTORE_FILE", $backupFile)."\n";
					
					dealWithError($msg, 0);
				}
			}
			
			# delete the backup file
			if (unlink($backupFile) ne 1)	# returns the number of files deleted (1 in our case)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $backupFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		# we are asked to die, so we will (with an error message)
		dealWithError($langMsgText->maketext("CANNOT_UPDATE_FILE", $fileToUpdate)."\n", 1);
	}
	
	return $status;
}


##############################################################################
# Removes record-types from an Entity file
sub removeRecordTypeFromEntity
{
	my ($recordTypesToRemove) = @_;
	
	my $line;
	my $data;
	if (!open(FH, $ftsEntityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsEntityFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	while ($line = <FH>)
	{
		$line =~ s/ //g;	# remove any spaces that maybe in the data of the line
		
		if ($line =~ m/^(.+=)/)
		{
			my $fromEntityRecordType = $1;
			
			$fromEntityRecordType =~ s/=//g;	# remove '=' from the end of the string
			
			# if this record-type (from the file) is in our remove list, then skip it
			my $foundIt = 0;
			my $recordType;
			foreach $recordType (@$recordTypesToRemove)
			{
				if ($recordType eq $fromEntityRecordType)
				{
					printAndLog $msgLabel_INFO.$langMsgText->maketext("REMOVE_FROM_ENTITY", $recordType)."\n";
					
					$foundIt = 1;
					last;
				}
			}
			
			if ($foundIt eq 0)
			{
				$data .= $line;
			}
		}
		
	}
	close FH;
	
	# Now write the changes to the file
	return safeFileUpdate($ftsEntityFile, $data, 0);
}


##############################################################################
# Adds record-types to an Entity file
sub addRecordTypeToEntity
{
	my ($recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $status;
	
	# first clean up the Enity file (so that the addition won't have any problems)
	$status = cleanupEntityFile($ftsEntityFile);
	
	if ($status)
	{
		my @data;
		if (!open(FH, $ftsEntityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsEntityFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			return 0;
		}
		@data = <FH>;
		close(FH);
		
		my $alreadyExit = 0;
		foreach my $addLine (@$recordTypesAndFieldsToAdd)
		{
			my ($addRT, $nu1) = split("=", $addLine);
			
			foreach my $fromLine (@data)
			{
				my ($fromRT, $nu2) = split("=", $fromLine);
				
				if ($fromRT eq $addRT)
				{
					$alreadyExit = 1;
					last;
				}
			}
		}
		
		if ($alreadyExit eq 0)
		{
			my $line;
			foreach $line (@$recordTypesAndFieldsToAdd)
			{
				$line .= "\n";
				push(@data, $line);
			}
			
			my $resultData;
			foreach $line (@data)
			{
				$resultData .= $line;
			}
			
			# Now write the changes to the file
			my $status = safeFileUpdate($ftsEntityFile, $resultData, $dieOnError);
			if ($status)
			{
				cleanupEntityFile($ftsEntityFile);
			}
		}
		else
		{
			dealWithError($langMsgText->maketext("ADD_ENTITY_EXIST")."\n", $dieOnError);
		}
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_ADD_RECORD", $ftsEntityFile)."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Adds record-types to the CQ search property file
sub addRecordTypeToProperty
{
	my ($recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $status;
	
	# first clean up the Enity file (so that the addition won't have any problems)
	$status = cleanupEntityFile($ftsEntityFile);
	
	my $tempPropertyFileName = $ftsCQPropertyFileName.".added-record-type.temp";
	
	if ($status)
	{
		$status = generateCQSearchPropertyFile($ftsDataDir.$ftsEntityFileName, $ftsDataDir.$tempPropertyFileName, 1);
	}
	
	if (($status eq 0) and ($dieOnError eq 1))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_ADD_RECORD", $ftsEntityFile)."\n", $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Removes record-types from the CQ search property file
sub removeRecordTypeFromCQSearchProperty
{
	my ($recordTypesToRemove) = @_;
	
	my @data;
	if (!open(FH, $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsCQPropertyFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		return 0;
	}
	@data = <FH>;
	close(FH);
	
	my $line;
	my $recordType;
	foreach $recordType (@$recordTypesToRemove)
	{
		my @mData;
		my $matched = 0;
		foreach $line ( @data )
		{
			if ($matched eq 0)
			{
				if ($line =~ m/(.*<CQEntity.*name="$recordType".*)/)
				{
					$matched = 1;
				}
			}
			else
			{
				if ($line =~ m/(.*<\/CQEntity>)/)
				{
					$matched = 0;
					$line = "";
				}
			}
			
			if ($matched eq 0)
			{
				push(@mData, $line);
			}
		}
		
		@data = @mData;
	}
	
	my $resultData = join("", @data);
	
	# Now write the changes to the file
	return safeFileUpdate($ftsCQPropertyFile, $resultData, 0);
}


##############################################################################
# Remove record-types from the Lucene index
sub removeRecordTypeFromIndex
{
	my ($recordTypesToRemove, $dieOnError) = @_;
	
	chdirAndValidate($ftsDataDir."Solr/exampledocs/", 1);
	
	my $java = $Q.$ftsJavaEXE.$Q;
	if ($ftsJREHomeDir eq "/")
	{
		# this means 'JRE_HOME' was not found on the systems envirement settnig
		$java = "java";
	}
	
	my $status = 1;
	
	my $recordType;
	foreach $recordType (@$recordTypesToRemove)
	{
		if ($status eq 0)
		{
			last;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("REMOVE_FROM_INDEX", $recordType)."\n";
		
		my @mustMatch = $langMsgText->maketext("REMOVE_RECORD_TYPE_FROM_IDEX");
		my $cmd = $java." -Ddata=args -Durl=http://".$ftsFtsServerName.":".$ftsFtsWASProfilePort."/solr/update -jar post.jar \"<delete><query>record_type:".$recordType."</query></delete>\"";
		
		$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, 0);
	}
	
	return $status;
}


##############################################################################
# Creates setup file for add-record-type operation
sub addRecordType_SetupFile
{
	my ($workingDir, $recordTypesAndFieldsToAdd, $dieOnError) = @_;
	
	my $entityFile = $workingDir."addRecordType_Entity.txt";
	my $status = 1;
	
	my $originalCQPropertyData;
	my $addRecordTypeCQPropertyData;
	
	# write the in-memory Entity data to file
	my $data;
	my $line;
	foreach $line (@$recordTypesAndFieldsToAdd)
	{
		$line = trim($line);
		$data .= $line."\n";
	}
	
	if (open(FH, ">".$entityFile))
	{
		print FH $data;
		close(FH);
	}
	else
	{
		$status = 0;
	}
	
	# create the temp CQ Search property file (we only care about the "<CQEntityList>" block from this file)
	if ($status)
	{
		$status = generateCQSearchPropertyFile($entityFile, $workingDir.$ftsCQPropertyFileName, 0);
		if ($status)
		{
			customizeFtsFiles(0, $workingDir.$ftsCQPropertyFileName, $entityFile, 0);
		}
	}
	
	if ($status)
	{
		# open and read the original CQ search property file
		if (open(FH, $ftsCQPropertyFile))
		{
			my $line;
			while ($line = <FH>)
			{
				$originalCQPropertyData .= $line;
			}
			close FH;
		}
		else
		{
			$status = 0;
		}
	}
	
	if ($status)
	{
		# open and read the 'add' record-type CQ search property file
		if (open(FH, $workingDir.$ftsCQPropertyFileName))
		{
			my $line;
			while ($line = <FH>)
			{
				$addRecordTypeCQPropertyData .= $line;
			}
			close FH;
		}
		else
		{
			$status = 0;
		}
	}
	
	#
	# at this point, both the 'add' and 'original' CQ property file are in memory
	#
	
	my $finalCQPropertyData = $originalCQPropertyData;
	my $finalAddCQPropertyData = $originalCQPropertyData;
	
	if ($status)
	{
		# extract "<CQEntityList>" from the 'add' record-type CQ search property file
		my $addBlock = "";
		if ($addRecordTypeCQPropertyData =~ /<CQEntityList>.+<\/CQEntityList>/sg)
		{
			$addBlock = $&;
			
			# some minor clean up
			$addBlock = trim($addBlock);
			$addBlock =~ s/^<CQEntityList>//;
			$addBlock =~ s/<\/CQEntityList>$//;
			$addBlock = "    ".trim($addBlock)."\n";
		}
		else
		{
			$status = 0;
		}
		
		if ($addBlock ne "")
		{
			# insert the new record-type(s) into the original property file
			my $addedCQEntityData = "";
			my $tempData = $originalCQPropertyData;
			
			if ($tempData =~ s/<CQEntityList>.+<\/CQEntityList>//s)
			{
				$addedCQEntityData = $&;
				
				# some minor clean up followed with add-record-type data post-fixing
				$addedCQEntityData = trim($addedCQEntityData);
				$addedCQEntityData =~ s/^<CQEntityList>//;
				$addedCQEntityData =~ s/<\/CQEntityList>$//;
				$addedCQEntityData = "    ".trim($addedCQEntityData)."\n".$addBlock;
				$addedCQEntityData = "<CQEntityList>\n".$addedCQEntityData."  <\/CQEntityList>";
				
				$finalCQPropertyData =~ s/<CQEntityList>.+<\/CQEntityList>/$addedCQEntityData/s;
			}
			else
			{
				$status = 0;
			}
			
			# add the new record-type(s) to the add-property file.
			$addedCQEntityData = "";
			$tempData = $originalCQPropertyData;
			
			if ($tempData =~ s/<CQEntityList>.+<\/CQEntityList>//s)
			{
				$addedCQEntityData = "<CQEntityList>\n".$addBlock."  <\/CQEntityList>";
				
				$finalAddCQPropertyData =~ s/<CQEntityList>.+<\/CQEntityList>/$addedCQEntityData/s;
			}
			else
			{
				$status = 0;
			}
		}
		else
		{
			$status = 0;
		}
	}
	
	#
	# if we are here with no error, it means we can write our 'add' and 'original' CQ property file
	#
	
	if ($status)
	{
		# now write the new 'add' CQ search property file to the file system
		if (open(FH, ">".$workingDir.$ftsCQPropertyFileName))
		{
			print FH $finalAddCQPropertyData;
			close(FH);
		}
		else
		{
			$status = 0;
		}
		
		# final step, add the added record-type to the original CQ property file
		if ($status eq 1)
		{
			$status = safeFileUpdate($ftsCQPropertyFile, $finalCQPropertyData, 0);
		}
	}
	
	return $status;
}


##############################################################################
# This function copies the CQ FTS source template for this deployment
# --copy_fts_template
sub Func_copy_fts_template
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FTS_TEMPLATE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("copy_fts_template");
	
	my $tmp = $ftsTargetDir;	# remember it so we can restore it
	
	printAndLog $msgLabel_INFO."mkpath -> [".$ftsTargetDir."]"."\n";
	File::Path::mkpath $ftsTargetDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("FTS_TEMPLATE_AT", $ftsTargetDir)."\n";
	
	$ftsTargetDir = $ftsDataDir;
	
	followDir($ftsSourceDir, \&copyDir);
	
	$ftsTargetDir = $tmp;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FTS_TEMPLATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function generates the Entity file
# --gen_entity_file
sub Func_gen_entity_file
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GENERATE_ENTITY", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("gen_entity_file");
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $entityParm = $parm." -o Entity";
	
	my @mustMatch = $langMsgText->maketext("GEN_ENTITY_FILE");
	executeExternalCommand($cmd.$entityParm, @mustMatch, $gDieKind_must_match, 1);
	validateFileGenerated($cmd.$entityParm, $ftsClearQuestHomeDir.$ftsEntityFileName, 1);
	
	if (!move($ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName))
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName).
						"\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GENERATE_ENTITY", $stat_DONE)."\n", 1;
}


##############################################################################
# This function collects data about the CQ FTS deployment for diagnostic
# --gather_diagnostic_data
sub Func_gather_diagnostic_data
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GATHER_DIAG_DATA", $stat_START)."\n", 1;
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	my $diagFolder = $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."-"."Diag-".$timeData."/";
	
	my $status;
	
	# make a copy of the files in "CQ.Search/SAMPL/"
	$status = copyFiles($ftsDataDir, $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/", 0);
	
	# make a copy of the files in "CQ.Search/<dbset>_<userdb>/Solr/sol/conf/"
	$status = copyFiles($ftsDataDir."Solr/solr/conf/", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/Solr/sol/conf/", 0);
	
	# make a copy of the files in "CQ.Search/<dbset>_<userdb>/logs/"
	$status = copyFiles($ftsDataDir."logs/", $diagFolder.$ftsDBSetName."_".$ftsUserDbName."/logs/", 0);
	
	# make a copy of "web.xml"
	my $webXMLFolder = findWebXMLFolder($ftsWASProfileName);
	$status = copyFiles($webXMLFolder, $diagFolder.substr($webXMLFolder, length($ftsRationalCommon)), 0);
	
	# make a copy of "server.xml"
	my $serverXMLFolder = findServerXMLFolder($ftsWASProfileName);
	$status = copyFiles($serverXMLFolder, $diagFolder.substr($serverXMLFolder, length($ftsRationalCommon)), 0);
	
	# make a copy of CQ Search profile logs
	my $cqsearchprofileLogsFolder = $ftsRationalCommon."CM/profiles/".$ftsWASProfileName."/logs/server1/";
	$status = copyFiles($cqsearchprofileLogsFolder, $diagFolder.substr($cqsearchprofileLogsFolder, length($ftsRationalCommon)), 0);
	
	$cqsearchprofileLogsFolder = $ftsRationalCommon."CM/profiles/".$ftsWASProfileName."/logs/";
	$status = copyFiles($cqsearchprofileLogsFolder, $diagFolder.substr($cqsearchprofileLogsFolder, length($ftsRationalCommon)), 0);
	
	# make a copy of CQ Search profile 'ffdc' if it exist
	my $cqsearchprofileLogsFolderFfdc = $ftsRationalCommon."CM/profiles/".$ftsWASProfileName."/logs/ffdc/";
	if (-e $cqsearchprofileLogsFolderFfdc)
	{
		$status = copyFiles($cqsearchprofileLogsFolderFfdc, $diagFolder.substr($cqsearchprofileLogsFolderFfdc, length($ftsRationalCommon)), 0);
	}
	
	# make a copy of CM profile logs
	my $cmprofileLogsFolder = $ftsRationalCommon."CM/profiles/"."cmprofile"."/logs/server1/";
	$status = copyFiles($cmprofileLogsFolder, $diagFolder.substr($cmprofileLogsFolder, length($ftsRationalCommon)), 0);
	
	$cmprofileLogsFolder = $ftsRationalCommon."CM/profiles/"."cmprofile"."/logs/";
	$status = copyFiles($cmprofileLogsFolder, $diagFolder.substr($cmprofileLogsFolder, length($ftsRationalCommon)), 0);
	
	# make a copy of CM profile 'ffdc' if it exist
	my $cmprofileLogsFolderFfdc = $ftsRationalCommon."CM/profiles/"."cmprofile"."/logs/ffdc/";
	if (-e $cmprofileLogsFolderFfdc)
	{
		$status = copyFiles($cmprofileLogsFolderFfdc, $diagFolder.substr($cmprofileLogsFolderFfdc, length($ftsRationalCommon)), 0);
	}
	
	# copy over any .txt, .trc, and .dmp files that are core dump files from FTS
	my $cqsearchprofileCrashFiles = $ftsRationalCommon."CM/profiles/".$ftsWASProfileName;
	copyFilesMatch($cqsearchprofileCrashFiles, $diagFolder.substr($cqsearchprofileCrashFiles, length($ftsRationalCommon)), ".txt", 0);
	copyFilesMatch($cqsearchprofileCrashFiles, $diagFolder.substr($cqsearchprofileCrashFiles, length($ftsRationalCommon)), ".dmp", 0);
	copyFilesMatch($cqsearchprofileCrashFiles, $diagFolder.substr($cqsearchprofileCrashFiles, length($ftsRationalCommon)), ".trc", 0);
	
	# copy over any .txt, .trc, and .dmp files that are core dump files from CM
	my $cmprofileCrashFiles = $ftsRationalCommon."CM/profiles/"."cmprofile";
	copyFilesMatch($cmprofileCrashFiles, $diagFolder.substr($cmprofileCrashFiles, length($ftsRationalCommon)), ".txt", 0);
	copyFilesMatch($cmprofileCrashFiles, $diagFolder.substr($cmprofileCrashFiles, length($ftsRationalCommon)), ".dmp", 0);
	copyFilesMatch($cmprofileCrashFiles, $diagFolder.substr($cmprofileCrashFiles, length($ftsRationalCommon)), ".trc", 0);
	
	if ($status eq 0)
	{
		dealWithError($langMsgText->maketext("DIAG_COLLECTION_ERROR")."\n", 0);
	}
	
	printAndLog $msgLabel_INFO."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DIAG_DATA_LOCATION", $diagFolder)."\n";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GATHER_DIAG_DATA", $stat_DONE)."\n", 1;
}


##############################################################################
# This function displays the headline of all scenarios, the headline and the 
# body of a scenario, or the headline and body of all scenarios
# --show_scenarios
sub Func_show_scenarios
{
	my $cqHomeDir = "";
	$cqHomeDir = $ENV{'CQ_HOME'};
	if (!defined($cqHomeDir) or $cqHomeDir eq "")
	{
		$cqHomeDir = $ENV{'CLEARqUEST_HOME'};
	}
	if ((substr($cqHomeDir, -1) ne "/") or (substr($cqHomeDir, -1) ne "\\"))
	{
		$cqHomeDir .= "/";
	}
	
	my $dataFile = $cqHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE")."-scenarios.xml";
	
	my $parser = new XML::Parser(ErrorContext => 3, Style => "Tree");
	my $tree;
	
	eval
	{
		$tree = $parser->parsefile($dataFile);
	};
	
	buildScenariosFromXMLFile("", $tree);
	
	my $idToShow;
	if (!defined($gShowScenarios) or ($gShowScenarios eq ""))
	{
		$idToShow = "headline";
	}
	else
	{
		$idToShow = $gShowScenarios;
	}
	
	$idToShow = lc($idToShow);
	
	foreach my $key (sort keys %gFtsScenarios)
	{
		my $value = $gFtsScenarios{$key};
		
		if (($idToShow eq "all") or ($idToShow eq "headline") or ($idToShow eq lc($value->{"id"})))
		{
			printAndLog "\n\n", 1;
			printAndLog $langMsgText->maketext("SCENARIO_ID", $value->{"id"}), 1;
			printAndLog "\n", 1;
			printAndLog $langMsgText->maketext("SCENARIO_HEADLINE", $value->{"headline"}), 1;
			printAndLog "\n\n", 1;
		}
		
		if (($idToShow eq "all") or $idToShow eq lc($value->{"id"}))
		{
			printAndLog $value->{"body"}, 1;
			printAndLog "\n\n", 1;
		}
	}
	
	printAndLog "\n", 1;
	
	if ($@)
	{
		# report any error that stopped parsing, or announce success
		$@ =~ s/at \/.*?$//s;
		my $err = $@;
		dealWithError($langMsgText->maketext("ERROR_IN_CFG_FILE", $dataFile)."\n".$err."\n", 1);
	}
}


##############################################################################
# This function will enable verbose output (always on for this release)
# --verbose
sub Func_verbose
{
	# no-op
	$gVerboseOutput = 1;
}


##############################################################################
# This function will scrub CQ oplogs (if multisite is not enabled)
# --scrub_oplog
sub Func_scrub_oplog
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SCRUB_OPLOG", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("scrub_oplog");
	
	if ($gIsWindows eq 1)
	{
		if ($gIsReplicated eq 0)
		{
			my @val = split(/\./, $cqVersion);
			if ($val[2] lt 1)		# check for 7.1.X, where X must be 1 or later
			{
				dealWithError($langMsgText->maketext("NEED_CQ_7_1_1")."\n", 1);
			}
			
			chdirAndValidate($ftsClearQuestHomeDir, 1);
			
			my @mustMatch = $langMsgText->maketext("SCRUB_OPLOG");
			my $cmd = "multiutil.exe scruboplog -dbset ".$ftsDBSetName." -family ".$ftsUserDbName." -user ".$ftsUserName." -password ".$ftsUserPassword." -before ".$gScrubOplogBeforeDate;
			executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 1);
		}
		else
		{
			dealWithError($langMsgText->maketext("WILL_NOT_SCRUB")."\n", 1);
		}
	}
	else
	{
		runMultiutilOnUnix();
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SCRUB_OPLOG", $stat_DONE)."\n", 1;
}


##############################################################################
# This function makes a backup copy of the CQ FTS deployment
# --backup_fts
sub Func_backup_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_BACKUP_CQFTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	$gBackupDestinationDir = File::Spec->rel2abs($gBackupDestinationDir);
	$gBackupDestinationDir =~ s/\\/\//sg;
	
	if (!(-d $gBackupDestinationDir))
	{
		dealWithError($langMsgText->maketext("WRONG_BACKUP_DESTINATION")."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
		# we are given a directory
	if ((substr($gBackupDestinationDir, -1) ne "/") or (substr($gBackupDestinationDir, -1) ne "\\"))
	{
		$gBackupDestinationDir .= "/";
	}
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	my $backupDirName = $gBackupDestinationDir.$ftsDBSetName."_".$ftsUserDbName.".Backup-".$timeData."/";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DISABLING_UPDATE_MODE")."\n";
	Func_stop_update_idx(1);
	
	my $tmp_ftsTargetDir = $ftsTargetDir;	# remember it so we can restore it
	my $tmp_ftsSourceDir = $ftsSourceDir;
	
	$ftsTargetDir = $backupDirName;
	$ftsSourceDir = $ftsDataDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKING_UP_TO", $backupDirName)."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKUP_TIME_DEPENDENCY")."\n";
	followDir($ftsSourceDir, \&copyDir);
	
	$ftsTargetDir = $tmp_ftsTargetDir;
	$ftsSourceDir = $tmp_ftsSourceDir;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("ENABLING_UPDATE_MODE")."\n";
	Func_start_update_idx(1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_BACKUP_CQFTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions archives the current CQ FTS depolyment
# --archive_fts
sub Func_archive_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ARCHIVE_CQFTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	# delete the WAS profile
	if (($ftsWASProfileName ne "cqsearchprofile") and (profileExist($ftsWASProfileName) eq 1))
	{
		Func_delete_fts_was_profile(1);
	}
	
	# disable CQ Web FTS
	Func_disable_cqweb_fts(1);
	
	# rename CQ FTS folder
	my $oldDirName = $ftsDataDir;
	
	if (substr($oldDirName, -1) eq "/")
	{
		$oldDirName = substr($oldDirName, 0, length($oldDirName) -1);
	}
	
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	my $timeData = sprintf "%04d.%02d.%02d-%02d.%02d.%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
	
	my $archivedDirName = $oldDirName.".Archived-".$timeData;
	printAndLog $msgLabel_INFO.$langMsgText->maketext("RENAMING_TO", $oldDirName, $archivedDirName)."\n";
	
	printAndLog $msgLabel_WARN.$langMsgText->maketext("LOGGER_STOP_FOR_RENAME")."\n", 1;
	closeLogger();
	
	if (!move($oldDirName, $archivedDirName))
	{
		# we can't recover or log on this error, so just die with an error
		print "\n".$gTerminationErrorMsg."\n";
		print "\n";
		die $langMsgText->maketext("FAILED_TO_MOVE_FOLDER", $oldDirName, $archivedDirName)."\n".$msgLable_oS_ERROR.$!."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO."\n";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("ARCHIVED_TO", $archivedDirName)."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ARCHIVE_CQFTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function removes the lock file 'write.lock', if one exist, from the 
# Lucene index
# --remove_lucene_idx_lock
sub Func_remove_lucene_idx_lock
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_LUCENE_IDX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	my $lockFile = $ftsLuceneIndexDir."write.lock";
	if (-f $lockFile)
	{
		# the lock file exist, delete it
		if (unlink($lockFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $lockFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IDX_IS_UnLOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IDX_IS_NOT_LOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_LUCENE_IDX", $stat_DONE)."\n", 1;
}


##############################################################################
# This locks the CQ FTS deployment so that any command line option that may 
# modify any setting is prevented
# --lock_cq_fts
sub Func_clear_state
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CLEAR_STATE", $stat_START)."\n", 1;
	
	if (-f $ftsStateFile)
	{
		# the state file exist, delete it
		if (unlink($ftsStateFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsStateFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("STATE_IS_CLEARED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("NO_STATE_TO_CLEAR")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CLEAR_STATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function locks the CQ FTS deployment so that any command line option 
# that may modify any setting is prevented.
# --lock_cq_fts
sub Func_lock_cq_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_LOCK_CQ_FTS", $stat_START)."\n", 1;
	
	if (!(-f $ftsLockFile))
	{
		if (!open(FH, ">$ftsLockFile"))
		{
			dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $ftsLockFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		print FH "";
		close FH;
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_LOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_ALREADY_LOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_LOCK_CQ_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function Unlocks the CQ FTS deployment so that any command line option 
# that may modify any setting is allowed.
# --unlock_cq_fts
sub Func_unlock_cq_fts
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_CQ_FTS", $stat_START)."\n", 1;
	
	if (-f $ftsLockFile)
	{
		# the lock file exist, delete it
		if (unlink($ftsLockFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsLockFile)."\n", 1);
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_UnLOCKED")."\n";
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CQFTS_IS_ALREADY_UNLOCKED")."\n";
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UNLOCK_CQ_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function validates if the CQ FTS deployment is valid or not
# --validate_cq_fts_deplo
sub Func_validate_cq_fts_deplo
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_VALIDATE", $stat_START)."\n", 1;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_VALIDATE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function will prepare a CQ FTS deployment for a CQ upgrade
# --prep_upgd_was_profiles
sub Func_prep_upgd_was_profiles
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_PREP_UPGD_PROFILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("prep_upgd_was_profiles");
	
	# make sure the backup directory exist, if not, create it
	if (!(-d $gPrepUpgdFtsWASProfiles))
	{
		printAndLog $msgLabel_INFO."mkpath -> [".$gPrepUpgdFtsWASProfiles."]"."\n";
		File::Path::mkpath $gPrepUpgdFtsWASProfiles;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BACKUP_PROFILES_TO", $gPrepUpgdFtsWASProfiles)."'\n";
	
	# get list of CQ FTS WAS profiles
	my (@profilesList) = getListOfProfiles();
	foreach my $item (@profilesList)
	{
		my @profileInfoData = ();
		my $backupPath = "";
		my $profileName = "";
		my $profilePort = "";
		my $profMaxHeapSize = "";
		my $serviceStartupType = "";
		my $genericJvmArg = "";
		my $ftsHome = "";
		my $webXMLFile = "";
		my $serverXMLFile = "";
		
		# get the profile name
		my $foundAt = rindex($item, "/");
		if ($foundAt ne -1)
		{
			$foundAt += 1;
			
			$profileName = substr($item, $foundAt);
			$backupPath = $gPrepUpgdFtsWASProfiles.$profileName."/";
			$backupPath =~ s/\\/\//sg;
		}
		
		# get the port number of this profile
		if ($profileName ne "")
		{
			$profilePort = findPortNumberForProfile($profileName);
		}
		
		# get the web.xml file
		if ($profilePort ne "")
		{
			$webXMLFile = findWebXMLFolder($profileName);
			if ($webXMLFile ne "")
			{
				$webXMLFile .= "web.xml";
				$webXMLFile =~ s/\\/\//sg;
			}
		}
		
		# get the server.xml file
		if ($webXMLFile ne "")
		{
			$serverXMLFile = findServerXMLFolder($profileName);
			if ($serverXMLFile ne "")
			{
				$serverXMLFile .= "server.xml";
				$serverXMLFile =~ s/\\/\//sg;
			}
		}
		
		# get the startup type
		if (($gIsWindows eq 1) and ($serverXMLFile ne ""))
		{
			$serviceStartupType = getServiceStartupType($profileName);
		}
		
		# get the max heap size (of the WAS profile) and generic JVM arguments
		if ($serverXMLFile ne "")
		{
			($genericJvmArg, $profMaxHeapSize, $ftsHome) = getHeapSizeAndJVMArg($serverXMLFile);
		}
		
		# if we collected all the data we need, write it to the backup data file
		if ($serverXMLFile eq "")
		{
			dealWithError($langMsgText->maketext("CANNOT_GATHER_PROFILE", $item, $profileName, $profilePort, $serverXMLFile, $webXMLFile)."\n", 0);
			next;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_BACKUP_OF_PROFILE", $profileName)."'\n";
		
		# make the directory for this profile's data
		printAndLog $msgLabel_INFO."mkpath -> [".$backupPath."]"."\n";
		File::Path::mkpath $backupPath;
		
		# delete any old state file (if any)
		unlink($backupPath."restore.state");
		
		# prepare data to write to the data file
		push(@profileInfoData, "FTS_HOME: ".$ftsHome);
		push(@profileInfoData, "PROF_NAME: ".$profileName);
		push(@profileInfoData, "PROF_PORT: ".$profilePort);
		push(@profileInfoData, "PROF_MAX_HEAP_SIZE: ".$profMaxHeapSize);		# from: server.xml
		push(@profileInfoData, "WIN_SERVICE_STARTUP_TYPE: ".$serviceStartupType);
		push(@profileInfoData, "GENERIC_JVM_ARG: ".$genericJvmArg);	# from: server.xml
		push(@profileInfoData, "WEB_XML: ".$webXMLFile);			# just restore the file
		push(@profileInfoData, "SERVER_XML: ".$serverXMLFile);		# don't restore this file, use set_solr_home and set_was_max_mem
		
		my $status = 1;
		
		# write the data
		my $dataFile = $backupPath.$profileName.".txt";
		eval
		{
			if (!open(FH, ">$dataFile"))
			{
				dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $dataFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
				$status = 0;
			}
			if ($status eq 1)
			{
				foreach my $item (@profileInfoData)
				{
					print FH $item."\n";
				}
				close FH;
			}
		};
		if ($@)
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $dataFile)."\n", 1);
			$status = 0;
		}
		
		# make a copy of web.xml and server.xml files into our backup directory
		if (($status eq 1) and !copy($webXMLFile, $backupPath))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $webXMLFile, $backupPath)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		if (($status eq 1) and !copy($serverXMLFile, $backupPath))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $serverXMLFile, $backupPath)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			$status = 0;
		}
		
		# skip 'cqsearchprofile'
		if (($profileName eq "cqsearchprofile") and (canDeleteDefaultProfile_cqsearchprofile() eq 0))
		{
			next;
		}
		
		# finaly, delete the WAS profile if they asked us to
		if (($status eq 1) and ($ftsDeleteFtsWASProfiles eq "true"))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FTS_WAS_PROFILE", $profileName)."'\n";
			deleteFtsWASProfile($profileName, 0);
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_PREP_UPGD_PROFILES", $stat_DONE)."\n", 1;
}


##############################################################################
# Restores CQ FTS WAS profiles
# --restore_was_profiles
sub Func_restore_was_profiles
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_RESTORE_PROFILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("restore_was_profiles");
	
	# make sure the backup directory exist, if not, create it
	if (!(-d $gRestoreFtsWASProfiles))
	{
		dealWithError($langMsgText->maketext("NO_SUCH_DIR_FILE", $gRestoreFtsWASProfiles)."\n", 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("RESTORING_PROFILES_FROM", $gRestoreFtsWASProfiles)."'\n";
	
	# get list of CQ FTS WAS profiles
	my (@profilesDirList) = getListOfProfilesDir($gRestoreFtsWASProfiles);
	foreach my $item (@profilesDirList)
	{
		my $backupDataFile = $gRestoreFtsWASProfiles.$item."/".$item.".txt";
		my $stateFile = $gRestoreFtsWASProfiles.$item."/restore.state";
		my $ftsHome;
		my $profileName = "";
		my $profilePort = "";
		my $profMaxHeapSize = "";
		my $serviceStartupType = "";
		my $genericJvmArg = "";
		my $webXMLFile = "";
		my $serverXMLFile = "";
		
		# if this profile is processed, move on to the next one
		if (-e $stateFile)
		{
			next;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("RESTORING_PROFILE", $backupDataFile)."'\n";
		
		if (!open(FH, $backupDataFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $backupDataFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			next;
		}
		
		my $line;
		while ($line = <FH>)
		{
			$line =~ s/\s+$//;
			
			if ($line =~ /FTS_HOME: /)
			{
				$ftsHome = $line;
				$ftsHome =~ s{FTS_HOME:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_NAME: /)
			{
				$profileName = $line;
				$profileName =~ s{PROF_NAME:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_PORT: /)
			{
				$profilePort = $line;
				$profilePort =~ s{PROF_PORT:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /PROF_MAX_HEAP_SIZE: /)
			{
				$profMaxHeapSize = $line;
				$profMaxHeapSize =~ s{PROF_MAX_HEAP_SIZE:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /WIN_SERVICE_STARTUP_TYPE: /)
			{
				$serviceStartupType = $line;
				$serviceStartupType =~ s{WIN_SERVICE_STARTUP_TYPE:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /GENERIC_JVM_ARG: /)
			{
				$genericJvmArg = $line;
				$genericJvmArg =~ s{GENERIC_JVM_ARG:\s(.*?)}{$1}x;
			}
			elsif ($line =~ /WEB_XML: /)
			{
				$webXMLFile = $line;
				$webXMLFile =~ s{WEB_XML:\s(.*?)}{$1}x;
				$webXMLFile =~ s/\\/\//sg;
			}
			elsif ($line =~ /SERVER_XML: /)
			{
				$serverXMLFile = $line;
				$serverXMLFile =~ s{SERVER_XML:\s(.*?)}{$1}x;
				$serverXMLFile =~ s/\\/\//sg;
			}
		}
		close FH;
		
		# skip 'cqsearchprofile'
		if (($profileName eq "cqsearchprofile") and (canDeleteDefaultProfile_cqsearchprofile() eq 0))
		{
			next;
		}
		
		# if we collected all the data we need, we can now restore the WAS profile
		if (($profileName eq "") or ($profilePort eq "") or ($genericJvmArg eq ""))
		{
			dealWithError($langMsgText->maketext("CANNOT_GATHER_PROFILE", $profileName, $profilePort, $genericJvmArg, $serverXMLFile, $serviceStartupType)."\n", 0);
			next;
		}
		
		my $status = 1;
		
		# recreate the WAS profile
		$status = createFtsWASProfile($ftsHome, $profileName, $profilePort, $serviceStartupType, 0);
		
		# restore the web.xml file
		if ($status eq 1)
		{
			my $fromWebXmlFile = $gRestoreFtsWASProfiles.$item."/"."web.xml";
			
			if (!copy($fromWebXmlFile, $webXMLFile))
			{
				dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $fromWebXmlFile, $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				$status = 0;
			}
		}
		
		# set WAS max memory
		if ($status eq 1)
		{
			if ($profMaxHeapSize ne "")
			{
				$status = startFtsWASProfile($profileName, 0);
				if ($status eq 1)
				{
					$status = setWASMaxMem($profileName, $profMaxHeapSize, 0);
				}
			}
		}
		
		# restart the WAS profile so all changes take effect
		if ($status eq 1)
		{
			$status = stopFtsWASProfile($profileName, 0);
			if ($status eq 1)
			{
				$status = startFtsWASProfile($profileName, 0);
			}
		}
		
		# set the state, for this profile that we are done
		if (!open(FH, ">$stateFile"))
		{
			dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $stateFile)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
		close FH;
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_RESTORE_PROFILES", $stat_DONE)."\n", 1;
}


##############################################################################
# This function deletes a CQ FTS WAS profile.
# --delete_fts_was_profile
sub Func_delete_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DELETE_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("delete_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = 0;
	
	if (($ftsWASProfileName eq "cqsearchprofile") and (canDeleteDefaultProfile_cqsearchprofile() eq 0))
	{
		dealWithError($langMsgText->maketext("CANNOT_DELETE_CQSEARCHPROFILE", $ftsWASProfileName)."\n", $dieOnError);
	}
	else
	{
		$status = deleteFtsWASProfile($ftsWASProfileName, $dieOnError);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DELETE_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This function creates a WAS profile.  Requires one of the following 
# parameter: manual, automatic, or disabled.  The parameter is only valid on 
# Windows.
# --create_fts_was_profile
sub Func_create_fts_was_profile
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CREATE_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("create_fts_was_profile");
	
	createFtsWASProfile($ftsDataDir, $ftsWASProfileName, $ftsFtsWASProfilePort, $gWinServiceStartUpType, 1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CREATE_WAS_PROFILE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function runs the repair-mode record extractor
# --repair_records
sub Func_repair_records
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REINDEX_RECORDS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("repair_records");
	
	if ($ftsBatchIndexJVMParm ne "")
	{
		$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
	}
	
	my $cmdPrefix = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -repair";
	my $cmdPostFix = $Q.$ftsCQPropertyFile.$Q." -site \"".$ftsSiteName."\"";
	
	if (!(-e $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("MISSING_CONFIG_FILE", $ftsCQPropertyFile)."\n", 1);
	}
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $file;
	my @files = <$ftsDataDir*.xml>;
	foreach $file (@files)
	{
		my $fileName;
		my $foundAt = rindex($file, "/") + 1;
		if ($foundAt ne -1)
		{
			$fileName = substr($file, $foundAt);
		}
		
		my $recordType;
		foreach $recordType (@$ftsListOfIndxedRecordTypes)
		{
			if ($fileName =~ m/^$recordType\d+\.xml/)
			{
				printAndLog $msgLabel_INFO.$langMsgText->maketext("RE_INDEXING_REPAIR", $file)."\n";
				
				my $cmd = $cmdPrefix." ".$Q.$file.$Q." ".$cmdPostFix;
				my @mustMatch = $langMsgText->maketext("REPAIR_RECORDS");
				my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 0);
				
				if ($status)
				{
					# delete the file
					if (unlink($file) ne 1)
					{
						dealWithError($langMsgText->maketext("CANNOT_DELETE", $file)."\n", 0);
					}
				}
			}
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REINDEX_RECORDS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function optimizes the index
# --optimize_idx
sub Func_optimize_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_OPTIMIZE_INDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("optimize_idx");
	
	if ($ftsBatchIndexJVMParm ne "")
	{
		$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
	}
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -optimize ".$Q.$ftsCQPropertyFile.$Q." -site \"".$ftsSiteName."\"";
	
	if (!(-e $ftsCQPropertyFile))
	{
		dealWithError($langMsgText->maketext("MISSING_CONFIG_FILE", $ftsCQPropertyFile)."\n", 1);
	}
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	my @mustMatch = $langMsgText->maketext("OPTIMIZE_IDX");
	executeExternalCommand($cmd, @mustMatch, $gDieKind_must_not_match, 1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_OPTIMIZE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function adds record-types
# --add_record_type
sub Func_add_record_type
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ADD_RECORD_TYPE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	# check and set state
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne "add_record_type"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "add_record_type", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, so set one to our state of 'setup_cq_fts'
		setState("add_record_type", "");
	}
	
	$gTopLevelSate = "add_record_type";
	
	# the data is in "record1=field1,field2,field3;record2=field1,field2" format so our splits are baed on ";", "=" and ",".
	my @rawRecordTypesAndFieldsToAdd = split(/;/, $ftsAddRecordType);
	
	if (scalar(@rawRecordTypesAndFieldsToAdd) eq 0)
	{
		dealWithError($langMsgText->maketext("NO_RECORD_TYPES_TO_ADD")."\n", 1);
	}
	
	##########################################################################
	
	my @cleanRecordTypesAndFieldsToAdd;
	
	foreach my $recordTypeBlocks (@rawRecordTypesAndFieldsToAdd)
	{
		my @recordTypeAndItsFields = split(/=/, $recordTypeBlocks);
		
		my $recordTypeName;
		my $fieldsList;
		
		if (scalar(@recordTypeAndItsFields) eq 2)
		{
			$fieldsList = $recordTypeAndItsFields[1];
		}
		
		$recordTypeName = $recordTypeAndItsFields[0];
		
		$recordTypeName = trim($recordTypeName);
		$fieldsList = cleanupFieldsList($fieldsList);
		
		isRecordTypeExist($recordTypeName, 1);
		isFieldListExist($recordTypeName, $fieldsList, 1);
		
		push(@cleanRecordTypesAndFieldsToAdd, $recordTypeName."=".$fieldsList);
	}
	
	##########################################################################
	
	# it's time to add record-type(s)
	my $status = 1;
	my $isIndexOk = 1;
	
	# create a restore poing of the files we are about to modify (in case we want to role back)
	# entity, property, intermediate, solr
	my @restoreFileList = ($ftsEntityFile, $ftsCQPropertyFile, $ftsSolrIntermediateFile, $ftsSolrFile);
	my $restoreExtension = ".restore.add_record_type";
	my $restoreFile;
	
	if (index($state, "restore_point") eq -1)
	{
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile, $restoreFile.$restoreExtension))
			{
				$status = 0;
				dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				last;
			}
		}
		
		if (! $status)
		{
			# we can't craete a restore point, so get out
			foreach $restoreFile (@restoreFileList)
			{
				if (unlink($restoreFile.$restoreExtension) ne 1)
				{
					dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
				}
			}
			
			# we have an error, so terminate
			dealWithError($langMsgText->maketext("ERROR_WHILE_REMOVING_RECORD")."\n", 1);
		}
		
		setDoneState("restore_point");
	}
	
	#
	# we have our restore point files in place, we are good to go
	#
	
	my $enableUpdateIdx = 0;
	
	# stop update mode
	if (($status) and (index($state, "stop_update_idx") eq -1))
	{
		Func_stop_update_idx(1);
		$enableUpdateIdx = 1;
		
		setDoneState("stop_update_idx");
	}
	
	##########################################################################
	# phase-1: add new record-type to: Entity, and schema.xml
	
	# add the new record-types to the existing Entity file
	if (($status) and (index($state, "add_to_entity") eq -1))
	{
		$status = addRecordTypeToEntity(\@cleanRecordTypesAndFieldsToAdd, 1);
		
		setDoneState("add_to_entity");
	}
	
	# re-generate intermediate Solr file
	if (($status) and (index($state, "add_to_intermediate_solr") eq -1))
	{
		$status = generateSolrIntermediateFile($ftsEntityFile, 0);
		
		setDoneState("add_to_intermediate_solr");
	}
	
	# customize the Solr file to include those new fields
	if (($status) and (index($state, "cust_solr_files") eq -1))
	{
		$status = Func_cust_solr_files(0, 1);
		
		setDoneState("cust_solr_files");
	}
	
	##########################################################################
	# phase-2: add new record-type to a specialy Entity, and CQ Search property file
	
	my $workingDir;
	
	$workingDir = $ftsDataDir."addRecordType.temp"."/";
	printAndLog $msgLabel_INFO."mkpath -> [".$workingDir."]"."\n";
	File::Path::mkpath $workingDir;
	
	# create a specialy Entity and CQ Search property file which will hold to-be-added record-types
	if (($status) and (index($state, "update_master_and_temp_property_file") eq -1))
	{
		$status = addRecordType_SetupFile($workingDir, \@cleanRecordTypesAndFieldsToAdd, 0);
		
		setDoneState("update_master_and_temp_property_file");
	}
	
	# restart WAS profile so that Solr picks up the changes to schema.xml
	if (($status) and (index($state, "restart_was_profile") eq -1))
	{
		$status = stopFtsWASProfile($ftsWASProfileName, 0);
		
		if ($status eq 1)
		{
			$status = startFtsWASProfile($ftsWASProfileName, 0);
			
			setDoneState("restart_was_profile");
		}
	}
	
	# send the new CQ search property file to CQ core 
	# (so that users can start searching on new record-types as they are added)
	if (($status) and (index($state, "enable_cqweb_fts") eq -1))
	{
		$status = Func_enable_cqweb_fts(0, 1);
		
		setDoneState("enable_cqweb_fts");
	}
	
	##########################################################################
	# phase-3: start indexing those new record-types
	
	# index newly added record-types
	if (($status) and (index($state, "run_batch_mode") eq -1))
	{
		if ($ftsBatchIndexJVMParm ne "")
		{
			$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
		}
		
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -batch ".$Q.$workingDir.$ftsCQPropertyFileName.$Q." -site "."\"".$ftsSiteName."\"";
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
		my @mustMatch = (": cqtsdbcrawer finished at ", ": cqtsdbcrawer  ", ": cqtsdbcrawer w_ ", ": cqtsdbcrawer  ");
		$status = executeExternalCommand($cmd, \@mustMatch, $gDieKind_must_match, 0);
		
		setDoneState("run_batch_mode");
	}
	
	##########################################################################
	
	# post record-type addition, deal with success or error
	if ($status)
	{
		# all is well, delete the restore files
		foreach $restoreFile (@restoreFileList)
		{
			if (unlink($restoreFile.$restoreExtension) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# delete the working directory (and the files in it)
		my @files = <$workingDir*>;
		foreach my $file (@files)
		{
			# clean up the temporary port file.
			if (unlink($file) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE", $file)."\n", 0);
			}
		}
		
		if (!rmdir($workingDir))
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE_FOLDER", $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	else
	{
		# some action failed, report an error (and terminate if the index was not modified, and it's OK)
		dealWithError($langMsgText->maketext("ERROR_DURING_ADD_RECORD")."\n", ! $isIndexOk);
		
		# restore from the backup
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile.$restoreExtension, $restoreFile))
			{
				dealWithError($langMsgText->maketext("FILE_RESTORE_ERROR", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# restore CQ Web search property
		Func_enable_cqweb_fts(0, 1);
		
		# if we run into an error and the index was modified, then we have an issue
		if (! $isIndexOk)
		{
			dealWithError($langMsgText->maketext("OUT_OF_SYNC_LUCENE_INDEX")."\n", $isIndexOk);
		}
	}
	
	##########################################################################
	
	# finally, re-start update mode (if all is well, or if we stopped earlier)
	if (($status or $enableUpdateIdx) and (index($state, "start_update_idx") eq -1))
	{
		Func_start_update_idx(1);	# start update mode
		
		setDoneState("start_update_idx");
	}
	
	setDoneState("");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ADD_RECORD_TYPE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function removes record-types
# --remove_record_type
sub Func_remove_record_type
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REMOVE_RECORD_TYPE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	##########################################################################
	
	# check and set state
	my ($function, $state) = getState();
	my $newState = 0;
	
	if (($function ne "") and ($function ne "remove_record_type"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "remove_record_type", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, set it ours
		setState("remove_record_type", "");
		$newState = 1;
	}
	
	$gTopLevelSate = "remove_record_type";
	
	##########################################################################
	
	# validate that record-type names are valid and exist for removal
	my @recordTypesToRemove = split(/;/, $ftsRemoveRecordType);
	
	if (scalar(@recordTypesToRemove) eq 0)
	{
		dealWithError($langMsgText->maketext("NO_RECORD_TYPES_TO_REMOVE")."\n", 1);
	}
	
	if ($newState eq 1)
	{
		# this is not a continuation of the state, so do a simple check
		foreach my $recordType (@recordTypesToRemove)
		{
			isThisRecordTypeIndexed($recordType, 1);
		}
	}
	
	##########################################################################
	
	my $status = 1;
	my $isIndexOk = 1;
	
	# create a restore point of files we are about to modify:
	# entity, property, intermediate, solr
	my @restoreFileList = ($ftsEntityFile, $ftsCQPropertyFile, $ftsSolrIntermediateFile, $ftsSolrFile);
	my $restoreExtension = ".restore.remove_record_type";
	my $restoreFile;
	
	if (index($state, "restore_point") eq -1)
	{
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile, $restoreFile.$restoreExtension))
			{
				$status = 0;
				dealWithError($langMsgText->maketext("CANNOT_CREATE_RESTORE_FILE", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
				last;
			}
		}
		
		if (! $status)
		{
			# we can't create a restore point, so get out
			foreach $restoreFile (@restoreFileList)
			{
				if (unlink($restoreFile.$restoreExtension) ne 1)
				{
					dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
				}
			}
			
			# we have an error, so terminate
			dealWithError($langMsgText->maketext("REMOVE_RECORD_ERROR")."\n", 1);
		}
		
		setDoneState("restore_point");
	}
	
	#
	# we have our restore point files in place, we are good to go
	# everything we do from here on, will end up modifying files and settings
	#
	
	my $enableUpdateIdx = 0;
	
	# stop update mode
	if (($status) and (index($state, "stop_update_idx") eq -1))
	{
		Func_stop_update_idx(1);
		$enableUpdateIdx = 1;
		
		setDoneState("stop_update_idx");
	}
	
	# remove it from the Entity file
	if (($status) and (index($state, "remove_from_entity") eq -1))
	{
		$status = removeRecordTypeFromEntity(\@recordTypesToRemove);
		
		setDoneState("remove_from_entity");
	}
	
	# remove it from the CQ search property file
	if (($status) and (index($state, "remove_from_cq_property") eq -1))
	{
		$status = removeRecordTypeFromCQSearchProperty(\@recordTypesToRemove);
		
		setDoneState("remove_from_cq_property");
	}
	
	# re-generate intermediate Solr file
	if (($status) and (index($state, "remove_from_intermediate_solr") eq -1))
	{
		$status = generateSolrIntermediateFile($ftsEntityFile, 0);
		
		setDoneState("remove_from_intermediate_solr");
	}
	
	# customize the Solr file to include those new fields
	if (($status) and (index($state, "cust_solr_files") eq -1))
	{
		$status = Func_cust_solr_files(0, 1);
		
		setDoneState("cust_solr_files");
	}
	
	# send the new CQ search property file to CQ core
	if (($status) and (index($state, "enable_cqweb_fts") eq -1))
	{
		$status = Func_enable_cqweb_fts(0, 1);
		
		setDoneState("enable_cqweb_fts");
	}
	
	# remove records of this record-type fromt the index
	if (($status) and (index($state, "remove_from_idx") eq -1))
	{
		$status = removeRecordTypeFromIndex(\@recordTypesToRemove, 0);	# delete it from Lucene's index
		
		$isIndexOk = $status;	# the index is OK or not
		
		setDoneState("remove_from_idx");
	}
	
	##########################################################################
	
	# post record-type removal, deal with success or error
	if ($status)
	{
		# all is well, delete the restore files
		foreach $restoreFile (@restoreFileList)
		{
			if (unlink($restoreFile.$restoreExtension) ne 1)
			{
				dealWithError($langMsgText->maketext("CANNOT_DELETE_RESTORE_FILE", $restoreFile.$restoreExtension)."\n", 0);
			}
		}
	}
	else
	{
		# some action failed, report an error (and terminate if the index was not modified, and it's OK)
		dealWithError($langMsgText->maketext("ERROR_DURING_REMOVE_RECORD")."\n", ! $isIndexOk);
		
		# restore from the backup
		foreach $restoreFile (@restoreFileList)
		{
			if (!copy($restoreFile.$restoreExtension, $restoreFile))
			{
				dealWithError($langMsgText->maketext("FILE_RESTORE_ERROR", $restoreFile, $restoreFile.$restoreExtension)."\n".$msgLable_oS_ERROR.$!."\n", 0);
			}
		}
		
		# restore CQ Web search property
		Func_enable_cqweb_fts(0, 1);
		
		# if we run into an error and the index was modified, then we have an issue
		if (! $isIndexOk)
		{
			dealWithError($langMsgText->maketext("OUT_OF_SYNC_LUCENE_INDEX")."\n", $isIndexOk);
		}
	}
	
	##########################################################################
	
	# finally, re-start update mode (if all is well, or if we stopped earlier)
	if (($status or $enableUpdateIdx) and (index($state, "start_update_idx") eq -1))
	{
		Func_start_update_idx(1);	# start update mode
		
		setDoneState("start_update_idx");
	}
	
	setDoneState("");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_REMOVE_RECORD_TYPE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function forces re-execution of the update mode as if it's being run
# for the first time.
# --fresh_update_idx
sub Func_fresh_update_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_UPDATE_REINDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("fresh_update_idx");
	
	my $theFile = $ftsDataDir."update_recovery.properties";
	
	if (-e $theFile)
	{
		stopFtsWASProfile($ftsWASProfileName, 1);
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_OPLOG_MARKER_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 1);
		}
		
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_UPDATE_REINDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function forces re-execution of the batch mode as if it's being run
# for the firest time.
# --fresh_batch_idx
sub Func_fresh_batch_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_BATCH_REINDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("fresh_batch_idx");
	
	my $theFile = $ftsDataDir."batch_recovery.properties";
	
	if (-e $theFile)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETE_BATCH_MARKER_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 1);
		}
	}
	
	Func_run_batch_idx();
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_FORCE_BATCH_REINDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function sets the JVM maximum memory of the WAS profile
# --set_was_max_mem
sub Func_set_was_max_mem
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_WAS_MAX_JVM_MEM", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	
	if ($ftsWasProfileMaxHeapSize ne "")
	{
		setWASMaxMem($ftsWASProfileName, $ftsWasProfileMaxHeapSize, 1);
	}
	else
	{
		# they are removing JVM setting, the only way to do this is to actualy modify server.xml file
		my $serverXMLFile = findServerXMLFolder($ftsWASProfileName);
		if ($serverXMLFile ne "")
		{
			$serverXMLFile .= "server.xml";
			$serverXMLFile =~ s/\\/\//sg;
			
			removeHeapSizeXMLTag($serverXMLFile); 
		}
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_WAS_MAX_JVM_MEM", $stat_DONE)."\n", 1;
}


##############################################################################
# This function will update the CQ property file
# --update_fts_prop_files
sub Func_update_fts_prop_files
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPDATE_FTS_FILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("update_fts_prop_files");
	
	customizeFtsFiles(1, $ftsCQPropertyFile, $ftsEntityFile, 1);		# this will sync the CQ Search property file with the CQ FTS config file
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_UPDATE_FTS_FILES", $stat_DONE)."\n", 1;
}


##############################################################################
# This function generates the Entity file (if one is not provided), the Solr
# intermediate file, and the CQ Search property file.
# --gen_fts_files
sub Func_gen_fts_files
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GEN_FTS_FILES", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("gen_fts_files");
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	my $entityFile;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	# If they provided an entity file, then don't generate one and use theirs instead
	if ((defined $ftsUserDefinedEntityFile) and (length($ftsUserDefinedEntityFile) != 0))
	{
		$entityFile = $ftsUserDefinedEntityFile;
		printAndLog $msgLabel_INFO.$langMsgText->maketext("USING_USER_ENTITY_FILE", $entityFile)."\n";
	}
	else
	{
		my $entityParm = $parm." -o Entity";
		my @mustMatch = $langMsgText->maketext("GEN_ENTITY_FILE");
		executeExternalCommand($cmd.$entityParm, @mustMatch, $gDieKind_must_match, 1);
		validateFileGenerated($cmd.$entityParm, $ftsClearQuestHomeDir.$ftsEntityFileName, 1);
		
		if (move($ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName) ne 1)
		{
			dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName).
							"\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		
		validateFileGenerated("move ".$ftsClearQuestHomeDir.$ftsEntityFileName." ".$ftsDataDir.$ftsEntityFileName, $ftsDataDir.$ftsEntityFileName, 1);
		
		$entityFile = $ftsEntityFile;
	}
	
	generateSolrIntermediateFile($entityFile, 1);
	generateCQSearchPropertyFile($entityFile, $ftsDataDir.$ftsCQPropertyFileName, 1);
	
	setDoneState("gen_fts_files");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_GEN_FTS_FILES", $stat_DONE)."\n", 1;
}


##############################################################################
# This function customizes the CQ Search property file
# --cust_fts_files
sub Func_cust_fts_files
{
	my ($isInUpdateMode) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_PROPERTY_FILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("cust_fts_files");
	
	if (!defined($isInUpdateMode))
	{
		$isInUpdateMode = 0;	# use default
	}
	
	customizeFtsFiles($isInUpdateMode, $ftsCQPropertyFile, $ftsEntityFile, 1);
	
	setDoneState("cust_fts_files");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_PROPERTY_FILE", $stat_DONE)."\n", 1;
}


##############################################################################
# This function customizes the Solr schema.xml file
# --cust_solr_files
sub Func_cust_solr_files
{
	my ($dieOnError, $ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_SOLR_FILE", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{
		dieIfCqFtsIsLocked();
		dieIfInState("cust_solr_files");
	}
	
	my $isFirstMatch = 1;
	my $line;
	my $data;
	
	my $status = 1;
	
	# Read the current schema.xml file
	if (open(FH, $ftsSolrFile))
	{
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		# Reflect the user-db name in schema.xml
		$data =~ s/<schema name=".+" version=".+">/<schema name="$ftsUserDbName" version="1.1">/;
		
		# Replace old <field>...</field> with new data based on user's DB
		my $solrFieldData = getSolrFieldData();    
		$data =~ s/<fields>.+<\/fields>/<fields>\n$solrFieldData<\/fields>/s;
		
		# Replace old <copyField>...</copyField> with new data based on user's DB
		my $solrCopyFieldData = getSolrCopyFieldData();
		
		if ($data =~ /(<copyField source=".+" dest=".+"><\/copyField>\n)+/smg)
		{
			# pre 7.1.0.2
			# <copyField source="Notes_Log" dest="RATL_CQ_ALL_FIELDS"></copyField>
			$data =~ s/(<copyField source=".+" dest=".+"><\/copyField>\n)+/$solrCopyFieldData/smg;
		}
		else
		{
			# post 7.1.0.2
			# <copyField dest="RATL_CQ_ALL_FIELDS" source="Action_Types"/>
			$data =~ s/(<copyField dest=".+" source=".+"\/>\n)+/$solrCopyFieldData/smg;
		}
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_SOLR_SCHEMA_FILE", $ftsSolrFile)."\n";
		
		$status = safeFileUpdate($ftsSolrFile, $data, $dieOnError);
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsSolrFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	setDoneState("cust_solr_files");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_CUST_SOLR_FILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This functions sends the 'start' command to the WAS profile
# --start_fts_was_profile
sub Func_start_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("start_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = startFtsWASProfile($ftsWASProfileName, $dieOnError);
	
	setDoneState("start_fts_was_profile");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This functions sends the 'stop' command to the WAS profile
# --stop_fts_was_profile
sub Func_stop_fts_was_profile
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_WAS_PROFILE", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("stop_fts_was_profile");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = stopFtsWASProfile($ftsWASProfileName, $dieOnError);
	
	setDoneState("stop_fts_was_profile");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_WAS_PROFILE", $stat_DONE)."\n", 1;
	
	return $status;
}


##############################################################################
# This function sets Solr's home environment via JVM's parameter
# --set_solr_home
sub Func_set_solr_home
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_SOLR_HOME", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("set_solr_home");
	
	setSolrHome($ftsWASProfileName, $ftsDataDir, 1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_SET_SOLR_HOME", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions enables CQ FTS by invoking the installutil command
# --enable_cqweb_fts
sub Func_enable_cqweb_fts
{
	my ($dieOnError, $ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ENABLE_CQWEB_FTS", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{
		dieIfCqFtsIsLocked();
		dieIfInState("enable_cqweb_fts");
	}
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	run_installutil(0, $dieOnError);
	
	setDoneState("enable_cqweb_fts");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_ENABLE_CQWEB_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This functions disables CQ FTS by invoking the installutil command
# --disable_cqweb_fts
sub Func_disable_cqweb_fts
{
	my ($dieOnError) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DISABLE_CQWEB_FTS", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("disable_cqweb_fts");
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	run_installutil(1, 1);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_DISABLE_CQWEB_FTS", $stat_DONE)."\n", 1;
}


##############################################################################
# This function executes the batch mode record extractor
# --run_batch_idx
sub Func_run_batch_idx
{
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_BATCH_INDEX", $stat_START)."\n", 1;
	
	dieIfCqFtsIsLocked();
	dieIfInState("run_batch_idx");
	
	if ($ftsDoBatchIndex eq "true")
	{
		if ($ftsBatchIndexJVMParm ne "")
		{
			$ftsBatchIndexJVMParm = " -jvmpar ".$Q.$ftsBatchIndexJVMParm.$Q;
		}
		
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q.$ftsBatchIndexJVMParm." -batch ".$Q.$ftsCQPropertyFile.$Q." -site "."\"".$ftsSiteName."\"";
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
		my @mustMatch = (": cqtsdbcrawer finished at ", ": cqtsdbcrawer  ", ": cqtsdbcrawer w_ ", ": cqtsdbcrawer  ");
		executeExternalCommand($cmd, \@mustMatch, $gDieKind_must_match, 1);
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_BATCH_INDEX")."\n", 1);
	}
	
	setDoneState("run_batch_idx");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_BATCH_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function enables the update mode record extractor
# --start_update_idx
sub Func_start_update_idx
{
	my ($ignoreCheck) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_UPDATE_INDEX", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{	
		dieIfCqFtsIsLocked();
		dieIfInState("start_update_idx");
	}
	
	if ($ftsStartupdateIndex eq "true")
	{
		my $webXMLFile = findWebXMLFolder($ftsWASProfileName);
		
		$webXMLFile .= "web.xml";
		
		$webXMLFile =~ s/\\/\//sg;
		
		my $line;
		my $data;
		if (!open(FH, $webXMLFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		my $newValue;
		
		$newValue = HTML::Entities::encode($ftsCQPropertyFile);
		$data =~ s/<param-value>\S*\s*\D*\d*CQ-\S*\s*\D*\d*.xml<\/param-value>/<param-value>$newValue<\/param-value>/s;
		$data =~ s{
			(<param-name>config</param-name>)	# $1 = "sitename" element
			(.*?)								# $2 = stuff between elements
			(<param-value>)						# $3 = param-value start tag
			.*?									# discard param-value content
			(</param-value>)					# $4 = param-value end tag
		}{$1$2$3$newValue$4}sgx;
		
		$newValue = HTML::Entities::encode($ftsSiteName);
		$data =~ s{
			(<param-name>sitename</param-name>) # $1 = "sitename" element
			(.*?)								# $2 = stuff between elements
			(<param-value>)						# $3 = param-value start tag
			.*?									# discard param-value content
			(</param-value>)					# $4 = param-value end tag
		}{$1$2$3$newValue$4}sgx;
		
		$newValue = "<load-on-startup>1</load-on-startup>";
		$data =~ s/<!--\s*<load-on-startup>1<\/load-on-startup>\s*-->/$newValue/sm;
		
		printAndLog $msgLabel_INFO.$langMsgText->maketext("MODIFYING_WEBXML_FILE", $webXMLFile)."\n";
		
		safeFileUpdate($webXMLFile, $data, 1);	# now update web.xml file
		
		stopFtsWASProfile($ftsWASProfileName, 1);
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_UPDATE_INDEX")."\n", 1);
	}
	
	setDoneState("");
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_START_UPDATE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# This function stops the update mode record extractor.
# --stop_update_idx
sub Func_stop_update_idx
{
	my ($ignoreCheck) = @_;

	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_UPDATE_INDEX", $stat_START)."\n", 1;
	
	if (!defined($ignoreCheck))
	{
		$ignoreCheck = 0;
	}
	
	if ($ignoreCheck eq 0)
	{	
		dieIfCqFtsIsLocked();
		dieIfInState("stop_update_idx");
	}
	
	my $webXMLFile = findWebXMLFolder($ftsWASProfileName);
	
	$webXMLFile .= "web.xml";
	
	$webXMLFile =~ s/\\/\//sg;
	
	my $line;
	my $data;
	if (!open(FH, $webXMLFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $webXMLFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
	}
	while ($line = <FH>)
	{
		$data .= $line;
	}
	close FH;
	
	# If it's allready disabled, there is nothing to do
	if (($data =~ /<!--\s*<load-on-startup>1<\/load-on-startup>\s*-->/) eq "")
	{
		my $newValue;
		
		$newValue = "\n    <!--\n    <load-on-startup>1</load-on-startup>\n    -->"."\n";
		$data =~ s/\s*<load-on-startup>1<\/load-on-startup>\s*/$newValue/sm;
		
		safeFileUpdate($webXMLFile, $data, 1);	# now update web.xml file
		
		stopFtsWASProfile($ftsWASProfileName, 1);
		startFtsWASProfile($ftsWASProfileName, 1);
	}
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("BANNER_STOP_UPDATE_INDEX", $stat_DONE)."\n", 1;
}


##############################################################################
# Deletes FTS WAS profile
sub deleteFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my $status = 1;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_WAS_PROFILE", $profileName)."'\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	printAndLog $msgLabel_INFO.$langMsgText->maketext("TIME_1_TO_5_MIN")."\n";
	printAndLog $msgLabel_INFO."************************************************************"."\n";
	
	chdirAndValidate($ftsRationalCommon."eWAS/bin/", $dieOnError);
	
	my @mustMatch = $langMsgText->maketext("DEL_WAS_PROFILE");
	my $cmd = $Q.$ftsRationalCommon."eWAS/bin/manageprofiles".$ext.$Q." -delete -profileName ".$profileName;
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FOLDERS_FILES", $ftsRationalCommon."CM/profiles/".$profileName)."\n";
	File::Path::rmtree($ftsRationalCommon."CM/profiles/".$profileName);
	
	return $status;
}


##############################################################################
# Creates FTS WAS profile
sub createFtsWASProfile
{
	my ($ftsHome, $profileName, $profilePort, $serviceStartupType, $dieOnError) = @_;
	
	my $status = 1;
	
	# make sure there is no such profiel alrady
	if (profileExist($profileName) eq 1)
	{
		dealWithError($langMsgText->maketext("WAS_PROFILE_EXIST", $profileName)."\n", $dieOnError);
		$status = 0;
	}
	
	# make srue the port is good
	if ($status eq 1)
	{
		if (length($profilePort) lt 5)
		{
			dealWithError($langMsgText->maketext("WAS_PORT_GT_5")."\n", $dieOnError);
			$status = 0;
		}
	}
	
	my $WASPropFile;
	
	# prepare the WAS property file (to create the WAS profile)
	if ($status eq 1)
	{
		my $wasProfileData = "WC_defaulthost=xxx080\nWC_adminhost=xxx060\nWC_defaulthost_secure=xxx443\nWC_adminhost_secure=xxx043\nBOOTSTRAP_ADDRESS=xxx809\nSOAP_CONNECTOR_ADDRESS=xxx880"."\n";
	
		my $port = substr($profilePort, 0, length($profilePort) - 3);
		
		$wasProfileData =~ s/=xxx/=$port/smg;
		
		eval
		{
			$WASPropFile = $ftsDataDir."cqftsportdef.props";
			printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_WAS_PROP_FILE", $WASPropFile)."\n";
			
			if (!open(FH, ">$WASPropFile"))
			{
				dealWithError("\n".$langMsgText->maketext("CANNOT_CREATE", $WASPropFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
				$status = 0;
			}
			print FH $wasProfileData;
			close FH;
		};
		if ($@)
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $WASPropFile)."\n", $dieOnError);
			$status = 0;
		}
	}
	
	# create the WAS profile
	if ($status eq 1)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_NEW_WAS_PROFILE", $profileName, $profilePort)."'\n";
		printAndLog $msgLabel_INFO."************************************************************"."\n";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("TIME_5_TO_10_MIN")."\n";
		printAndLog $msgLabel_INFO."************************************************************"."\n";
		
		chdirAndValidate($ftsRationalCommon."eWAS/bin/", 1);
		
		# Create a new WAS profile
		my $extraWindowsOpt = "";
		if ($gIsWindows eq 1)
		{
			$extraWindowsOpt = "-winserviceCheck true -winserviceStartupType ".$serviceStartupType;
		}
		else
		{
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_SRV_TYPE", $serviceStartupType)."\n", 1;
		} 
		
		my $cmd = $Q.$ftsRationalCommon."eWAS/bin/manageprofiles".$ext.$Q." -create -templatePath ".$Q.$ftsRationalCommon."eWAS/profileTemplates/default".$Q." -profileName ".$profileName." -profilePath ".$Q.$ftsRationalCommon."CM/profiles/".$profileName.$Q." ".$extraWindowsOpt." -nodeName ".$profileName." -hostName localhost -cellName dfltCell -portsFile ".$Q.$WASPropFile.$Q." -omitAction samplesInstallAndConfig defaultAppDeployAndConfig";
		
		my @mustMatch = $langMsgText->maketext("CRE_WAS_PROFILE");
		executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	}
	
	# clean up the temporary port file.
	if (-e $WASPropFile)
	{
		if (unlink($WASPropFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $WASPropFile)."\n", 0);
		}
	}
	
	# now deploy Solr
	if ($status eq 1)
	{
		my $solrWarFile = $ftsRationalCommon."CM/apps/apache-solr-1.2.0.war";
		
		if (!(-e $solrWarFile))
		{
			# CQ 7.1.0.1 or ealier
			$solrWarFile = $ftsClearQuestHomeDir."CQ.Solr/SAMPL/Solr/apache-solr-1.2.0/dist/apache-solr-1.2.0.war";
		}
		
		$status = createDeplySolrScrip(0);
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -conntype NONE -lang jython -profileName ".$profileName." -f ".$Q.$ftsDataDir."deploySolr.py".$Q." ".$Q.$solrWarFile.$Q." server1 solr";
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminConfig save".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_SOLR_SAVE");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
	}
	
	# Depoly CqTsDbCrawler
	if ($status eq 1)
	{
		my $crawlerWarFile = $ftsRationalCommon."CM/apps/cqtsdbcrawler.war";
		
		if (!(-e $crawlerWarFile))
		{
			$crawlerWarFile = $ftsClearQuestHomeDir."cqtsdbcrawler.war";
		}
		
		if (-e $crawlerWarFile)
		{
			my $cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminApp install {".$crawlerWarFile."} { -contextroot cqtsdbcrawler -appname cqtsdbcrawler -MapWebModToVH {{cqtsdbcrawler cqtsdbcrawler.war,WEB-INF/web.xml default_host}}}".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_CRAWLER");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
		else
		{
			$status = 0;
		}
		
		if ($status eq 1)
		{
			my $cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -conntype NONE -profileName ".$profileName." -c ".$Q."\$AdminConfig save".$Q;
			my @mustMatch = $langMsgText->maketext("DEPLOY_CRAWLER_SAVE");
			$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
		}
	}
	
	# delete Solr deploy script
	deleteDeplySolrScrip();
	
	# Set Solr's home
	if ($status eq 1)
	{
		$status = setSolrHome($profileName, $ftsHome, $dieOnError);
	}
	
	return $status;
}


##############################################################################
# Set the JVM maximum memory of the WAS profile
sub setWASMaxMem
{
	my ($profileName, $maxHeapSize, $dieOnError) = @_;
	
	my $cmd = $ftsRationalCommon."CM/scripts/";
	
	my $status = chdirAndValidate($cmd, $dieOnError);
	
	$cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -profile -profileName ".$profileName." ".$Q.$cmd."teamAdminUtils.jacl".$Q." -c ".$Q."setJvmMaxHeapSize ".$maxHeapSize.$Q;
	
	my @mustMatch = $langMsgText->maketext("SET_WAS_MAX_MEM");
	$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Prevent execution if the FTS deployment is locked
sub dieIfCqFtsIsLocked
{
	if (-f $ftsLockFile)
	{
		dealWithError("\n".$langMsgText->maketext("CQFTS_IS_LOCKED_ERROR")."\n", 1);
	}
}


##############################################################################
# Terminates if we are trying to execute a command-line option that will alter
# the current state
sub dieIfInState
{
	my ($cmd) = @_;
	
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne $gTopLevelSate))
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("IN_A_STATE_ALREADY", $cmd, $function)."\n";
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
}


##############################################################################
# Runs the CQ installutil command (to enable / disable CQ FTS)
sub run_installutil
{
	my ($doRemove, $dieOnError) = @_;
	
	if (!defined($dieOnError))
	{
		$dieOnError = 1;	# use default
	}
	
	my $status = 1;
	
	if ($ftsIsEnableCQFTS eq "true")
	{
		if ($gIsWindows eq 1)
		{
			### ###
			#
			# ENABLING:
			# ---------
			# will enable at all sites or the single non-replicated DB only
			# installutil.exe setmasterpropertyfromfile TextSearch -user_db_name SAMPL admin "" TEXT_SEARCH_PROPERTY "C:/CQ.Search/TextSearch_SAMPL/CQ-TextSearch-SAMPL.xml"
			#
			# will enable at a site (replicated only) 
			# installutil.exe setmasterpropertyfromfile TextSearch -user_db_name SAMPL admin "" TEXT_SEARCH_PROPERTY "C:/CQ.Search/TextSearch_SAMPL/CQ-TextSearch-SAMPL.xml" -site "site-name"
			#
			# DISABLING:
			# ----------
			# will disable for ALL sites (removes everything) or the single non-replicated DB only
			# installutil.exe setmasterpropertyfromfile TextSearch -user_db_name SAMPL admin "" TEXT_SEARCH_PROPERTY "C:/CQ.Search/TextSearch_SAMPL/CQ-TextSearch-SAMPL.xml" -allsites -remove
			#
			# will disable for a site (replicated) or the single non-replicated DB only
			# installutil.exe setmasterpropertyfromfile TextSearch -user_db_name SAMPL admin "" TEXT_SEARCH_PROPERTY "C:/CQ.Search/TextSearch_SAMPL/CQ-TextSearch-SAMPL.xml" -site "site-name" -remove
			#
			
			my @siteList = ();
			my $siteString = "";
			my $removeString = "";
			my $runCmd = 1;
			
			if ($ftsIsReplicated eq 1)
			{
				# complex case, we are replicated
				
				if ($ftsIsSiteWorkingMaster eq 0)
				{
					my $die = 1;
					my $env = $ENV{"CQ_AT_WORKING_NO_DIE"};
					if (defined($env))
					{
						$die = 0;
					}
					dealWithError($langMsgText->maketext("REQUIRES_AT_WORKING_MASTER")."\n", $die);
					$runCmd = 0;
				}
				else
				{
					if (($ftsEnabledAtReplicaNamesList eq "") or 
						compareArrays($ftsEnabledAtReplicaNamesList, $ftsReplicaNamesList) eq 1)
					{
						# we are enabling / disabling CQ FTS at all sites
						if ($doRemove eq 1)
						{
							$removeString = " -allsites -remove";
						}
					}
					else
					{
						# we are enabling / disabling CQ FTS at some sites
						$siteString = " -site ";
						if ($doRemove eq 1)
						{
							$removeString = " -remove";
						}
					}
					
					if ($siteString eq "")
					{
						# enable FTS for all sites
						
						push(@siteList, "");
					}
					else
					{
						# enable FTS at some sites
						$siteString = " -site ";
						@siteList = split(",", $ftsEnabledAtReplicaNamesList);
					}
				}
			}
			else
			{
				# simple case, we are not replicated
				if ($doRemove eq 1)
				{
					$removeString = " -allsites -remove";
				}
				push(@siteList, "");
			}
			
			if ($runCmd eq 1)
			{
				chdirAndValidate($ftsClearQuestHomeDir, $dieOnError);
				
				my @mustMatch = $langMsgText->maketext("INSTALLUTIL");
				
				# execuite installutil.exe 
				foreach my $siteName (@siteList)
				{
					my $siteData = "";
				
					if ($siteName ne "")
					{
						$siteData = $siteString.$siteName." ";
					}
					
					my $cmd = "installutil.exe setmasterpropertyfromfile ".$ftsDBSetName." -user_db_name ".$ftsUserDbName." ".$ftsUserName." ".$ftsUserPassword." TEXT_SEARCH_PROPERTY ".$Q.$ftsCQPropertyFile.$Q.$siteData.$removeString;
					$status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
				}
			}
		}
		else
		{
			runInstallUtilOnUnix();
			
			# if we are in a stateful command, terminate the current state, and present message on what need to be done on a Windows machine
			my ($function, $state) = getState();
			if ($function ne "")
			{
				if ($doRemove eq 0)
				{
					setDoneState("enable_cqweb_fts");
				}
				else
				{
					setDoneState("disable_cqweb_fts");
				}
				
				stopInstallUtilWithWindowsMsg();
				
				closeLogger();	# close the logger
				
				exit 0;			# exit script, all is well
			}
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("NOT_ENABLED_CQWEB_FTS")."\n", 1);
	}
	
	return $status;
}


##############################################################################
# Customizes FTS files (for now, CQ Search properties file only)
sub customizeFtsFiles
{
	my ($isInUpdateMode, $cqProperityFile, $cqEntityFile, $recreatePW) = @_;
	
	my $data;
	my $line;
	
	# If a password is given, then generate an encrypted version of it, and a key file for it
	if (($ftsUserPassword ne "\"\"") and ($recreatePW eq 1))
	{
		my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCQtsdbcrawlerPL.$Q." -encrypt ";
		my $parm = " -u ".$ftsUserName." -p ".$ftsUserPassword." -s ".$ftsDBSetName." -d ".$ftsUserDbName." -kf ".$Q.$ftsKeyFile.$Q." -pf ".$Q.$ftsPWDFile.$Q;
		
		chdirAndValidate($ftsClearQuestHomeDir, 1);
        
		$cmd .= $parm;
		
		my @mustMatch = $langMsgText->maketext("MATCH_NOTHING");
		executeExternalCommand($cmd, @mustMatch, $gDieKind_none, 0);
		validateFileGenerated($cmd, $ftsKeyFile, 1);
		validateFileGenerated($cmd, $ftsPWDFile, 1);
		
		# Open the CQ Search property file and add to it the key and pwd info
		
		my $pwd;
		
		if (!open(FH, $ftsPWDFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsPWDFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$pwd .= $line;
		}
		close FH;
		
		if (!open(FH, $cqProperityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cqProperityFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		while ($line = <FH>)
		{
			$data .= $line;
		}
		close FH;
		
		$data =~ s/keyfile=".*?"/keyfile="$ftsKeyFile"/sg;
		$data =~ s/password=".*?"/password="$pwd"/sg;
		
		safeFileUpdate($cqProperityFile, $data, 1);
	}
    
	$data = "";
    
    #
	# Open the CQ Search property file and change the crawler setting based on new input (if any) from the user
	#
	
	my @lines;
	{
		if (!open(FH, $cqProperityFile))
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $cqProperityFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		@lines = <FH>;
		close( FH );
	}
	
	my ($nu1, $nu2) = getEntityFileWithoutDispFieldInfo($cqEntityFile, 0, 1);	# all that we want is the list of display fields
	
	my $currRecordType = "";
	
	foreach $line ( @lines )
	{
		# change the size attribute of batch element lines if necessary
		if ($line =~ /^\s*<batch\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsBatchIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsBatchIndexDelay"/;
			$line =~ s/\bthread="\d+"/thread="$ftsBatchIndexThreads"/;
		}
		
		# change the size attribute of update element lines if necessary
		if ($line =~ /^\s*<update\s/)
		{
			$line =~ s/\bsize="\d+"/size="$ftsUpdateIndexBatchSize"/;
			$line =~ s/\bdelay="\d+"/delay="$ftsUpdateIndexDelay"/;
			$line =~ s/\blogin_iteration="\d+"/login_iteration="$ftsUpdateIndexLoginInterval"/;
		}
		
		# change the port and FTS server name
		if (($line =~ /^\s*<searchrequest\s/) or ($line =~ /^\s*<batch\s/) or ($line =~ /^\s*<update\s/))
		{
			$line =~ s/:\d*080\//:$ftsFtsWASProfilePort\//;
			$line =~ s/\/\/.*?:/\/\/$ftsFtsServerName:/sg;
		}
		
		# Set the "dbid" field as the default display field
		if (($isInUpdateMode eq 0) and (($line =~ /^\s*<field\s/) and ($line =~ /name="dbid"/)))
		{
			$line =~ s/disp_order="0"/disp_order="9999"/;
		}
		
		# set the display-field to whatever the user want it to be (set via the Entity file)
		unless (not %gDispOrdFieldList)
		{
			if ($line =~ m/(.*<CQEntity.*name=.*)/)
			{
				while (my ($recordType, $field) = each (%gDispOrdFieldList))
				{
					if ($line =~ m/(.*<CQEntity.*name="$recordType".*)/)
					{
						$currRecordType = $recordType;
					}
				}
			}
			if ($line =~ m/(.*<\/CQEntity>)/)
			{
				$currRecordType = "";
			}
			if ($currRecordType ne "")
			{
				while (my ($recordType, $field) = each (%gDispOrdFieldList))
				{
					if ($currRecordType eq $recordType)
					{
						if ($line =~ /^\s*<field\s/)
						{
							# we matched a field line
							if ($line =~ /name="$field"/)
							{
								# we matched out field name, change it to be the display field
								$line =~ s/disp_order="0"/disp_order="1"/;
							}
							else
							{
								# it's not our display field, reset it to '0' in case if it was set to '1'
								$line =~ s/disp_order="1"/disp_order="0"/;
							}
						}
					}
				}
			}
		}
		
		# workaround known defect: RATLC01276003
		if (($cqVersion eq "7.1") or ($cqVersion eq "7.1.0") or ($cqVersion eq "7.1.0.0"))
		{
			my $locked_by_field = "<field name=\"locked_by\" disp_order=\"0\" index=\"true\"></field>";
			if ($line =~ s/$locked_by_field//)
			{
				$line = "";
			}
		}
		
		$data .= $line;
	}
	
	safeFileUpdate($cqProperityFile, $data, 1);
}


##############################################################################
# This function tries to create a new deployment of CQ FTS; it checks to see 
# if one already exists at the current location, if not then it will create
# the directory and push the 2 functions to be executed to complete the 
# operation.  The 2 functions are: Func_copy_fts_template() and 
# Func_gen_entity_file()
# --init_cq_fts
sub initCqFTS
{
	dieIfCqFtsIsLocked();
	dieIfInState("init_cq_fts");
	
	if (!(-f $optCqFtsHomeDir) and !(-d $optCqFtsHomeDir))
	{
		# the location they want to initalize CQ FTS at is not a config file or an existing 
		# folder; this is an initial setup so create our inital folders and files
		createNewCqFtsSetup();
	}
	elsif (-d $optCqFtsHomeDir)
	{
		# yes, folder exist, go after CQ FTS XML config file
		
		my $optCqFtsHomeDir_org = $optCqFtsHomeDir;
		
		$optCqFtsHomeDir = findExistingCqFtsSetup();	# try whereever the user asked us to look
		if ($optCqFtsHomeDir eq "")
		{
			# we can't find our CQ FTS config file, so revert back to new setup
			$optCqFtsHomeDir = $optCqFtsHomeDir_org;
			createNewCqFtsSetup();
		}
		else
		{
			my $dieMsg = $langMsgText->maketext("FTS_ALREADY_EXIST", $ftsDBSetName, $ftsUserDbName, $ftsTargetDir.$ftsDBSetName."_".$ftsUserDbName."/");
			dealWithError("\n".$dieMsg."\n", 1);
		}
	}
	else
	{
		my $dieMsg = $langMsgText->maketext("CANNOT_BE_HERE");
		dealWithError("\n".$dieMsg."\n", 1);
	}
	
	push(@gCmdToExecute, ["Func_"."copy_fts_template"]);		# --copy_fts_template
	push(@gCmdToExecute, ["Func_"."gen_entity_file"]);			# --gen_entity_file
}


##############################################################################
# Returns the state we are in
sub getState
{
	if (-f $ftsStateFile)
	{
		my $data;
		
		# the file exist, read it and return the data
		if (open(FH, $ftsStateFile))
		{
			my $line;
			
			while ($line = <FH>)
			{
				$data .= $line;
			}
			close FH;
		}
		
		if (!defined($data))
		{
			$data = "=";
		}
		
		my ($function, $state) = split(/=/, $data);
		
		$function = trim($function);
		$state = trim($state);
		
		return ($function, $state);
	}
	else
	{
		return ("", "");
	}
}


##############################################################################
# Sets the state we should be in
sub setState
{
	my ($function, $state) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("SET_STATE", $function, $state)."\n";
	
	my ($functionOld, $stateOld) = getState();
	
	if (($function eq "") and ($state eq ""))
	{
		# the state is being reset to nothing, so just delete the state file
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CLEARING_STATE")."\n";
		if ((-f $ftsStateFile) and (unlink($ftsStateFile) ne 1))
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $ftsStateFile)."\n", 1);
		}
	}
	else
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CURRENT_STATE", $function, $state)."\n";
	
		# the state for the current function is being change
		if (($functionOld ne "") and ($function ne $functionOld))
		{
			# somehow, we are trying to overwrite an existing state, this is an error
			dealWithError($langMsgText->maketext("UNEXPECTED_STATE_ERROR", $function, $functionOld)."\n", 1);
		}
		
		if ($stateOld ne "")
		{
			$stateOld .= ",";
		}
		
		$stateOld .= $state;
		
		my $data = $function."=".$stateOld;
		
		if (open(FH, ">$ftsStateFile"))
		{
			print FH $data;
			close FH;
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_WRITE_TO_FILE", $ftsStateFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Completes the state transaction
sub setDoneState
{
	my ($state) = @_;
	
	if ($gTopLevelSate eq "setup_cq_fts")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("setup_cq_fts", $state);
		}
	}
	elsif ($gTopLevelSate eq "remove_record_type")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("remove_record_type", $state);
		}
	}
	elsif ($gTopLevelSate eq "add_record_type")
	{
		if ($state eq "")
		{
			setState("", "");	# this means we are done, so reset the state
		}
		else
		{
			setState("add_record_type", $state);
		}
	}
}


##############################################################################
# Checks to see if we are in a state or not
sub isInState
{
	my ($function, $state) = getState();
	
	if ($function ne "")
	{
		return 0;
	}
	else
	{
		return 1;
	}
}


##############################################################################
# Pushes the required functions to perform end-to-end setup of CQ FTS.  It 
# also sets the 'state' file so that re-execution of the this function (in 
# the event of a failure) knowns where to continue from.
# --setup_cq_fts
sub setupCQFts
{
	# 1) if the state file doesn't exist, create it
	# 2) if the state file exist, based on it's state push functions
	
	dieIfCqFtsIsLocked();
	
	my ($function, $state) = getState();
	
	if (($function ne "") and ($function ne "setup_cq_fts"))
	{
		# we are in a state, and it's not our state
		dealWithError($langMsgText->maketext("IN_A_STATE_ALREADY", "setup_cq_fts", $function)."\n", 0);
		dealWithError($langMsgText->maketext("CLEAR_STATE_FOR", $function, "clear_state")."\n", 1);
	}
	elsif ($function eq "")
	{
		# we are not in any state, so set one to our state of 'setup_cq_fts'
		setState("setup_cq_fts", "");
	}
	elsif ($function eq "setup_cq_fts")
	{
		# we are in our state for which we need to continue 
		setState("setup_cq_fts", "");
	}
	
	$gTopLevelSate = "setup_cq_fts";
	
	# we are going to do a default end-to-end setup
	
	if (index($state, "gen_fts_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."gen_fts_files"]);				# --gen_fts_files
	}
	if (index($state, "cust_fts_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."cust_fts_files", 0]);			# --cust_fts_files
	}
	if (index($state, "cust_solr_files") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."cust_solr_files", 1]);			# --cust_solr_files
	}
	if (index($state, "stop_fts_was_profile") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."stop_fts_was_profile", 0]);		# --stop_fts_was_profile
	}
	if (index($state, "set_solr_home") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."set_solr_home"]);				# --set_solr_home
	}
	if (index($state, "start_fts_was_profile") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."start_fts_was_profile", 1]);		# --start_fts_was_profile
	}
	if (index($state, "enable_cqweb_fts") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."enable_cqweb_fts", 1]);			# --enable_cqweb_fts
	}
	if (index($state, "run_batch_idx") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."run_batch_idx"]);				# --run_batch_idx
	}
	if (index($state, "start_update_idx") eq -1)
	{
		push(@gCmdToExecute, ["Func_"."start_update_idx"]);				# --start_update_idx
	}
}


##############################################################################
# Starts a WAS profile
sub startFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my @mustMatch = $langMsgText->maketext("START_WAS_PROFILE");
	my $cmd = $Q.$ftsRationalCommon."eWAS/bin/startServer".$ext.$Q." server1 -profileName ".$profileName;
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Stops a WAS profile
sub stopFtsWASProfile
{
	my ($profileName, $dieOnError) = @_;
	
	my @mustMatch = $langMsgText->maketext("STOP_WAS_PROFILE");
	my $cmd = $Q.$ftsRationalCommon."eWAS/bin/stopServer".$ext.$Q." server1 -profileName ".$profileName;
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Sets Solr's home directory
sub setSolrHome
{
	my ($profileName, $dataDir, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsRationalCommon."eWAS/bin/wsadmin".$ext.$Q." -conntype NONE -lang jython -profileName ".$profileName." -f ".$Q.$ftsSetSolrHomePY.$Q." -interactive:no ".$Q.$dataDir."Solr/solr".$Q;
	
	my @mustMatch = $langMsgText->maketext("SET_SOLR_HOME");
	my $status = executeExternalCommand($cmd, @mustMatch, $gDieKind_must_match, $dieOnError);
	
	return $status;
}


##############################################################################
# Instructions on running installutil on Windows
sub stopInstallUtilWithWindowsMsg
{
	my ($function, $state) = getState();
	
	printAndLog $msgLabel_WARN.""."\n", 1;
	printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_ENABLE_FTS")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_BEFORE_CONTINUE")."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_SEE_FILE", $installUtilMsgFile)."\n", 1;
	printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_ONCE_COMPLETE", $function)."\n", 1;
	printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
	printAndLog $msgLabel_WARN.""."\n", 1;
}


##############################################################################
# Output file about running installutil on Windows
sub runInstallUtilOnUnix
{
	my $installUtilOnUnixData = $langMsgText->maketext("UNIX_INSTALLUTIL_INFO",
									$ftsDBSetName, $ftsCQPropertyFile, $optCqFtsHomeDir, 
									$ftsUserName, $ftsDBSetName, $ftsUserDbName, $ftsTargetDir, 
									$ftsCQPropertyFile);
	
	if ($gIsWindows eq 0)
	{
		if (!(-f $installUtilMsgFile))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("UNIX_INSTALLUTIL_MSG_FILE", $installUtilMsgFile)."\n";
		
			if (open(FH2, ">:utf8", $installUtilMsgFile))
			{
				print FH2 $installUtilOnUnixData;
				close FH2;
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_CREATE", $installUtilMsgFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
			}
		}
		
		if ($ftsIsEnableCQFTS eq "true")
		{
			printAndLog $msgLabel_WARN.""."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_ENABLE_FTS")."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_SEE_FILE", $installUtilMsgFile)."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_HOW_TO_CQWEB_FTS")."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.""."\n", 1;
		}
	}
}


##############################################################################
# Output file about running multiutil on Windows
sub runMultiutilOnUnix
{
	my $multiutilOnUnixData = $langMsgText->maketext("UNIX_MULTIUTIL_INFO", 
									$ftsDBSetName, $ftsCQPropertyFile, $optCqFtsHomeDir, 
									$ftsUserName, $ftsDBSetName, $ftsUserDbName, $ftsTargetDir);
	
	if ($gIsWindows eq 0)
	{
		if (!(-f $multiutilMsgFile))
		{
			printAndLog $msgLabel_INFO.$langMsgText->maketext("UNIX_MULTIUTIL_MSG_FILE", $multiutilMsgFile)."\n";
		
			if (open(FH2, ">:utf8", $multiutilMsgFile))
			{
				print FH2 $multiutilOnUnixData;
				close FH2;
			}
			else
			{
				dealWithError($langMsgText->maketext("CANNOT_CREATE", $multiutilMsgFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
			}
		}
		
		if ($ftsIsEnableCQFTS eq "true")
		{
			printAndLog $msgLabel_WARN.""."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_NOT_SUPPORTED_SCRUB")."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_SEE_FILE", $multiutilMsgFile)."\n", 1;
			printAndLog $msgLabel_WARN.$langMsgText->maketext("UNIX_HOW_TO_SCRUB")."\n", 1;
			printAndLog $msgLabel_WARN."**************************************************************"."\n", 1;
			printAndLog $msgLabel_WARN.""."\n", 1;
		}
	}
}


##############################################################################
# Finds the web.xml folder of a WAS profile
sub findWebXMLFolder
{
	my ($profileName) = @_;
	
	my $webXMLFolder = $ftsRationalCommon."CM/profiles/".$profileName."/config/cells/";
	
	if (index($webXMLFolder, ' ') ne -1)
	{
		$webXMLFolder = $Q.$webXMLFolder.$Q;
	}
	
	my $cellDir;
	my @files = <$webXMLFolder*>;
	foreach $cellDir (@files)
	{
		if ($cellDir =~ /Cell/)
		{
			$webXMLFolder = $cellDir;
			last;
		}
	}
	
	$webXMLFolder .= "/applications/cqtsdbcrawler.ear/deployments/cqtsdbcrawler/cqtsdbcrawler.war/WEB-INF/";
	
	return $webXMLFolder;
}


##############################################################################
# Finds the server.xml folder of a WAS profile
sub findServerXMLFolder
{
	my ($profileName) = @_;
	
	my $cellsFolder = $ftsRationalCommon."CM/profiles/".$profileName."/config/cells/";
	
	if (index($cellsFolder, ' ') ne -1)
	{
		$cellsFolder = $Q.$cellsFolder.$Q;
	}
	
	my $cellDir;
	my @files = <$cellsFolder*>;
	foreach $cellDir (@files)
	{
		if ($cellDir =~ /Cell/)
		{
			$cellsFolder = $cellDir;
			last;
		}
	}
	
	my $serverXMLFile = $cellsFolder."/nodes/";
	
	if (index($serverXMLFile, ' ') ne -1)
	{
		$serverXMLFile = $Q.$serverXMLFile.$Q;
	}
	
	my $nodeDir;
	my $myNodeDir = "";
	@files = <$serverXMLFile*>;
	foreach $nodeDir (@files)
	{
		if ($nodeDir =~ /Node/)
		{
			$myNodeDir = $nodeDir;
			last;
		}
	}
	
	if ($myNodeDir eq "")
	{
		$serverXMLFile .= $profileName."/servers/server1/";		# try our best
	}
	else
	{	
		$serverXMLFile = $myNodeDir."/servers/server1/";
	}

	$serverXMLFile =~ s/$Q//sg;
	if (!(-d $serverXMLFile))
	{
		$serverXMLFile = "";
	}
	
	return $serverXMLFile;
}


##############################################################################
# Finds out the service start up type (Windows only)
sub getServiceStartupType
{
	my ($profileName) = @_;
	
	my $foundStartupType = "manual";	# default to 'manual'
	
	if ($gIsWindows eq 1)
	{
		my $cmd = "sc qc \"IBMWAS61Service - ".$profileName."\"";	# sc qc "IBMWAS61Service - xyz"
		
		# execute the command
		if (open(FH, "$cmd 2>&1 |"))
		{
			my $line;
			
			while ($line = <FH>)
			{
				if ($line =~ /START_TYPE/)
				{
					if ($line =~ /AUTO_START/)
					{
						$foundStartupType = "automatic";
					}
					elsif ($line =~ /DEMAND_START/)
					{
						$foundStartupType = "manual";
					}
					elsif ($line =~ /DISABLED/)
					{
						$foundStartupType = "disabled";
					}
				}
			}
			close FH;
		}
	}
	
	return $foundStartupType;
}


##############################################################################
# Finds out the JVM heap size and the parameter
sub getHeapSizeAndJVMArg
{
	my ($serverXMLName) = @_;
	
	my $genericJvmArg = "";
	my $maxHeapSize = "";
	my $ftsHome = "";
	
	$serverXMLName =~ s/$Q//sg;
	
	if (open(FH, $serverXMLName))
	{
		my $line;
		while ($line = <FH>)
		{
			$line = trim($line);
			
			if ($line =~ /genericJvmArguments/)
			{
				# matching: <jvmEntries ... ... genericJvmArguments="XXXXX" ... .../>
				$genericJvmArg = $line;
				$genericJvmArg =~ s{
					<jvmEntries.*?
					genericJvmArguments
					.*?
					=.*?"(.*?)"
					.*?
					/>
				}{$1}x;
				
				# matching: -Dsolr.solr.home=C:/CQ.Search/TextSearch_SAMPL/Solr/solr
				$genericJvmArg =~ s{
					(-Dsolr.solr.home=)(.*?)
				}{$2}x;
				
				# matching: C:/CQ.Search/TextSearch_SAMPL/Solr/solr
				$ftsHome = $genericJvmArg;
				$ftsHome =~ s{
					(.*?)(Solr/solr)
				}{$1}x;
				
				$ftsHome =~ s/\\/\//sg;
				if (substr($ftsHome, -1) ne "/")
				{
					$ftsHome .= "/";
				}
				
				# remove any extras that are not part of 'solr.solr.home'
				my $foundAt = rindex($ftsHome, " ");
				if ($foundAt ne -1)
				{
					$ftsHome = substr($ftsHome, 0, $foundAt);
				}
			}
			
			if ($line =~ /maximumHeapSize/)
			{
				# matching: <jvmEntries ... ... maximumHeapSize="XXXXX" ... .../>
				$maxHeapSize = $line;
				$maxHeapSize =~ s{
					<jvmEntries.*?
					maximumHeapSize
					.*?
					=.*?"(.*?)"
					.*?
					/>
				}{$1}x;
			}
		}
		close FH;
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $serverXMLName)."\n", 0);
	}
	
	return ($genericJvmArg, $maxHeapSize, $ftsHome);
}


##############################################################################
# Removes the heap size setting from the server.xml file
sub removeHeapSizeXMLTag
{
	my ($serverXMLName) = @_;
	
	my $status = 1;		# assume we will not fail
	
	if (open(FH, $serverXMLName))
	{
		my $serverXMLData = "";
		my $wasChanged = 0;
		
		my $line;
		while ($line = <FH>)
		{
			if ($line =~ /maximumHeapSize/)
			{
				# matching: <jvmEntries ... ... maximumHeapSize="XXXXX" ... .../>
				# to remove: maximumHeapSize="XXXXX"
				$line =~ s{
					(<jvmEntries)(.*?)
					(maximumHeapSize)(.*?)(=)(.*?)(")(.*?)(")
					(.*?)(/>)
				}{$1$2$10$11}x;
				
				$wasChanged = 1;
			}
			
			$serverXMLData .= $line;
		}
		close FH;
		
		if ($wasChanged eq 1)
		{		
			$status = safeFileUpdate($serverXMLName, $serverXMLData, 0);
		}
	}
	else
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $serverXMLName)."\n", 0);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Creates the Solr deployment (under WAS) script
sub createDeplySolrScrip
{
	my ($dieOnError) = @_;
	
	my $status = 1;
	
	my $scriptData = <<EOF;

#
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2008, 2009.  All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

import sys, java

#-----------------------------------------------------------------
# Deploy Solr
#-----------------------------------------------------------------
def deploySolr(appLocation, serverName, contextRoot):

   global AdminApp
   global AdminConfig

   print "deploySolr, running with: "
   print "	Application Location: " + appLocation
   print "	Server name: " + serverName

   appArgs = "[-server " + serverName + " -contextroot " + contextRoot + " -appname apache-solr-1.2.0 -MapWebModToVH [[apache-solr-1.2.0.war apache-solr-1.2.0.war,WEB-INF/web.xml default_host]]]"

   AdminApp.install(appLocation, appArgs)
     
   AdminConfig.save()
   #the server should be restarted before trying to use the datasource


#-----------------------------------------------------------------
# Main
#-----------------------------------------------------------------
if (len(sys.argv) != 3):
   print "deploySolr: this script requires 3 parameters: " 
   print "	  1) the name of the node on which the application will be deployed,"
   print "	  2) the name of the server under which to create resources,"
   print "	  3) The contextRoot,"

   print ""
   print "e.g.:     deploySolr C:/myApps/myApp1.war server1 contextRoot"  
else:
   deploySolr(sys.argv[0], sys.argv[1], sys.argv[2])
EOF
	
    my $theFile;
    
	eval
	{
		$theFile = $ftsDataDir."deploySolr.py";
		printAndLog $msgLabel_INFO.$langMsgText->maketext("WAS_PROFILE_DEPOLY_FILE", $theFile)."\n";
		
		if (!open(FH, ">$theFile"))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
			$status = 0;
		}
		print FH $scriptData;
		close FH;
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $theFile)."\n", $dieOnError);
		$status = 0;
	}
	
	return $status;
}


##############################################################################
# Deletes the Solr deploy file
sub deleteDeplySolrScrip
{
	my $theFile = $ftsDataDir."deploySolr.py";
	
	if (-e $theFile)
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("DELETING_FILE", $theFile)."\n";
		if (unlink($theFile) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $theFile)."\n", 0);
		}
	}
}


##############################################################################
# Creates the CQ FTS admin configuration file
sub createFTSConfigXMLFile
{
	my ($configFile) = @_;
	
	my $configFileName = $ftsClearQuestHomeDir."CQTsAdmin/L10N/".$langMsgText->maketext("LANGUAGE_CODE").".xml";
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("READING_FTS_CONFIG_XML_FILE", $configFileName)."\n";
	
	if (!open FH, $configFileName)
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $configFileName)."\n", 1);
	}
	my @temp = <FH>;
	close FH;
	
	my $configData = join("",@temp);
	
	my $creationDate = "".localtime();
	$configData =~ s/\[date\]/$creationDate/s;
	
	# my $configData = $langMsgText->maketext("CONFIG_FILE_DATA", "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>", "".localtime());
	my $entityFile = $ftsDataDir."Entity-".$ftsDBSetName."-".$ftsUserDbName.".txt";
	my $siteName = "";
	
	if (($ftsIsReplicated eq 1) and ($ftsLocalReplicaName ne "") and ($ftsLocalReplicaName ne "<local>"))
	{
		$siteName = HTML::Entities::decode($ftsLocalReplicaName);
	}
	
	$configData =~ s/\[destination\]/$ftsTargetDir/s;
	$configData =~ s/\[dbSet\]/$ftsDBSetName/s;
	$configData =~ s/\[userDB\]/$ftsUserDbName/s;
	$configData =~ s/\[localhost\]/$ftsFtsServerName/s;
	$configData =~ s/\[entity-file\]/$entityFile/s;
	$configData =~ s/\[siteName\]/$siteName/s;
	
	if ($optCreateCqFtsWasProfile eq "1")
	{
		my $wasProfileNamePrefix = "cqsearchprofile_";
		
		# we are being asked for a new profile creation during initial setup
		$ftsWASProfileName = $wasProfileNamePrefix.$ftsDBSetName."_".$ftsUserDbName;
		if (length($ftsWASProfileName) gt $gMaxWASProfileLen)
		{
			my $oldName = $ftsWASProfileName;
			
			$wasProfileNamePrefix = "cqfts_";	# try a shorter name
			$ftsWASProfileName = $wasProfileNamePrefix.$ftsDBSetName."_".$ftsUserDbName;
			if (length($ftsWASProfileName) gt $gMaxWASProfileLen)
			{
				dealWithError($langMsgText->maketext("PROFILE_NAME_TOO_LONG", $ftsWASProfileName, $gMaxWASProfileLen)."\n", 1);
			}
			else
			{
				printAndLog $msgLabel_WARN.$langMsgText->maketext("WAS_PROFILE_NAME_TOO_LONG", $gMaxWASProfileLen, $ftsWASProfileName, $oldName)."\n", 1;
			}
		}
		
		$ftsFtsWASProfilePort = findNextPortNumberForProfile();
		
		$configData =~ s/<newValue required="no">cqsearchprofile_\[db-set-name\]_\[user-db-name\]<\/newValue>/<newValue required="no">$ftsWASProfileName<\/newValue>/s;
		$configData =~ s/<newValue required="no">\[\*\]<\/newValue>/<newValue required="no">$ftsFtsWASProfilePort<\/newValue>/s;
	}
	else
	{
		# give them the default (i.e.: use the default profile and port)
		$configData =~ s/<newValue required="no">cqsearchprofile_\[db-set-name\]_\[user-db-name\]<\/newValue>/<newValue required="no"><\/newValue>/s;
		$configData =~ s/<newValue required="no">\[\*\]<\/newValue>/<newValue required="no"><\/newValue>/s;
	}
	
	eval
	{
		printAndLog $msgLabel_INFO.$langMsgText->maketext("CREATING_FTS_CONFIG_XML_FILE", $configFile)."\n";
		
		if (!open(FH, ">$configFile"))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $configFile)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
		print FH $configData;
		close FH;
	};
	if ($@)
	{
		dealWithError($langMsgText->maketext("CANNOT_CREATE", $configFile)."\n", 1);
	}
}


##############################################################################
# Gets the copy-field data off the generated Solr file
sub getSolrCopyFieldData
{
	my $line;
	my @lines;
	my $data = "";
	
	if (!open FH, $ftsSolrIntermediateFile)
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $ftsSolrIntermediateFile)."\n", 1);
	}
	@lines = <FH>;
	close FH;
	
	foreach $line ( @lines )
	{
		if ($line =~ /<copyField source=".+" dest="RATL_CQ_ALL_FIELDS"><\/copyField>/)
		{
			$data .= $line;
		}
	}
	
	if ($data eq "")
	{
		# we are in 7.1.1 and later
		foreach $line ( @lines )
		{
			if ($line =~ /<copyField dest="RATL_CQ_ALL_FIELDS" source=".+"\/>/)
			{
				$data .= $line;
			}
		}
	}
	
	return $data;
}


##############################################################################
# Gets the fields data from Solr's schema.xml file
sub getSolrFieldData
{
	my $line;
	my @lines;
	my $data = "";
	
	open FH, $ftsSolrIntermediateFile;
	@lines = <FH>;
	close FH;
	
	foreach $line ( @lines )
	{
		if ($line =~ /<field name=".+" type="(text|string)" indexed="(true|false)" stored="(true|false)" required="(true|false)" multiValued="(true|false)"><\/field>/)
		{
			$data .= $line;
		}
	}
	
	if ($data eq "")
	{
		# we are in 7.1.1 and later
		foreach $line ( @lines )
		{
			if ($line =~ /<field indexed="(true|false)" multiValued="(true|false)" name=".+" required="(true|false)" stored="(true|false)" type="(text|string)"\/>/)
			{
				$data .= $line;
			}
		}
	}
	
	return $data;
}


##############################################################################
# Reads CQ FTS admin configuration file
sub readConfigData
{
	my ($dataFile) = @_;
	
	printAndLog $msgLabel_INFO.$langMsgText->maketext("READING_FTS_CONFIG_XML_FILE", $dataFile)."\n";
	
	my $parser = XML::Parser->new(ErrorContext => 3);
	
	$parser->setHandlers(	Start	=>	\&startElement,
							End		=>	\&endElement,
							Char	=>	\&characterData	);
	
	eval
	{
		$parser->parsefile($dataFile);
	};
	
	if ($@)
	{
		# report any error that stopped parsing, or announce success
		$@ =~ s/at \/.*?$//s;
		my $err = $@;
		dealWithError($langMsgText->maketext("ERROR_IN_CFG_FILE", $dataFile)."\n".$err."\n", 1);
	}
}


##############################################################################
# Creates an Entity file without the '&' in it (if the source has it)
sub getEntityFileWithoutDispFieldInfo
{
	my ($entityFile, $createCleanEntityFile, $dieOnError) = @_;
	
	if (!open(FH, $entityFile))
	{
		dealWithError($langMsgText->maketext("CANNOT_OPEN", $entityFile)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
	}
	
	my @newEntityData;
	
	my $line;
	while ($line = <FH>)
	{
		# Project=ratl_mastership,ratl_keysite,dbid,is_active,version,lock_version,record_type,&Name,Description,
		if ($line =~ /&/)
		{
			my $foundAt = 0;
			
			# get the record-type name
			$foundAt = index($line, "=");
			if ($foundAt ne -1)
			{
				my $recordTypeName;
				my $fieldName;
				
				$recordTypeName = trim(substr($line, 0, $foundAt));
				
				# now get the list of fields
				$foundAt = 0;
				while ($foundAt ne -1)
				{
					$foundAt = index($line, "&", $foundAt);
					if ($foundAt ne -1)
					{
						$foundAt = $foundAt + 1;
						
						my $endAt = index($line, ",", $foundAt);
						if ($endAt eq -1)
						{
							$fieldName = trim(substr($line, $foundAt));
						}
						else
						{
							$fieldName = trim(substr($line, $foundAt, $endAt - $foundAt));
						}
						
						$gDispOrdFieldList{$recordTypeName} = $fieldName;
					}
				}
				
				$line =~ s/&//g;
				
				push(@newEntityData, $line);
			}
		}
		else
		{
			# this is a record-type with no display field setting
			my $foundAt = index($line, "=");
			if ($foundAt ne -1)
			{
				$gDispOrdFieldList{trim(substr($line, 0, $foundAt))} = "";
				
				push(@newEntityData, $line);
			}
		}
	}
	
	close FH;
	
	my $entityFileClean;
	my $isNew = 0;
	
	if ((@newEntityData ne 0) and ($createCleanEntityFile eq 1))
	{
		$entityFileClean = $entityFile."-clean";
	
		if (!open(FH, ">".$entityFileClean))
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		}
		
		foreach my $line (@newEntityData)
		{
			print FH $line;
		}
		
		close FH;
		
		$entityFile = $entityFileClean;
		$isNew = 1;
	}
	
	return ($entityFile, $isNew);
}


##############################################################################
# Generates Solr intermediate file
sub generateSolrIntermediateFile
{
	my ($entityFile, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $status = 1;
	
	my ($entityFileClean, $isUseNewFile) = getEntityFileWithoutDispFieldInfo($entityFile, 1, $dieOnError);
	
	my $solrParm = $parm." -o Solr -f ".$entityFileClean;
	
	my @mustMatch = $langMsgText->maketext("GEN_INTER_SOLR_FILE");
	executeExternalCommand($cmd.$solrParm, @mustMatch, $gDieKind_must_match, $dieOnError);
	validateFileGenerated($cmd.$solrParm, $ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $dieOnError);
	if (move($ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $ftsDataDir.$ftsSolrIntermediateFileName) ne 1)
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE", $ftsClearQuestHomeDir.$ftsSolrIntermediateFileName, $ftsDataDir.$ftsSolrIntermediateFileName)."\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	if ($isUseNewFile eq 1)
	{
		# delete the temporary 'clean' Entity file
		if (unlink($entityFileClean) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	
	return $status;
}


##############################################################################
# Generates CQ Search property file
sub generateCQSearchPropertyFile
{	
	my ($entityFile, $moveTo, $dieOnError) = @_;
	
	my $cmd = $Q.$ftsCQPerl.$Q." ".$Q.$ftsCqtssetupPL.$Q." ";
	my $parm = "-s ".$ftsDBSetName." -d ".$ftsUserDbName." -u ".$ftsUserName." -p ".$ftsUserPassword;
	
	chdirAndValidate($ftsClearQuestHomeDir, 1);
	
	my $status = 1;
	
	my ($entityFileClean, $isUseNewFile) = getEntityFileWithoutDispFieldInfo($entityFile, 1, $dieOnError);
	
	my $cqParm = $parm." -o CQ -f ".$entityFileClean;
	
	my @mustMatch = $langMsgText->maketext("GEN_CQ_PROP_XML_FILE");
	executeExternalCommand($cmd.$cqParm, @mustMatch, $gDieKind_must_match, $dieOnError);
	validateFileGenerated($cmd.$cqParm, $ftsClearQuestHomeDir.$ftsCQPropertyFileName, $dieOnError);
	if (move($ftsClearQuestHomeDir.$ftsCQPropertyFileName, $moveTo) ne 1)
	{
		dealWithError($langMsgText->maketext("FAILED_TO_MOVE_FILE_TO", $ftsClearQuestHomeDir.$ftsCQPropertyFileName, $moveTo).
						"\n".$msgLable_oS_ERROR.$!."\n", $dieOnError);
		$status = 0;
	}
	
	if ($isUseNewFile eq 1)
	{
		# delete the temporary 'clean' Entity file
		if (unlink($entityFileClean) ne 1)
		{
			dealWithError($langMsgText->maketext("CANNOT_DELETE", $entityFileClean)."\n".$msgLable_oS_ERROR.$!."\n", 0);
		}
	}
	
	return $status;
}


##############################################################################
# Builds a scenarios data map for scenarios help
my %scenarios = ();
sub buildScenariosFromXMLFile
{
	my ($tag, $current) = @_;
	
	my $type;
	for (my $k = 0; $k < @$current; $k++)
	{
		if ($type = ref $$current[$k])
		{
			if ($type eq "ARRAY")
			{
				buildScenariosFromXMLFile($tag, $$current[$k]);
			}
			elsif ($type eq "HASH")
			{
				my $temp = $$current[$k];
				if (scalar keys %$temp ne 0)
				{
					die "\n\n\nFATAL ERROR: should not be here!! -- displayhash()\n\n";
				}
			}
		}
		else
		{
			my $info = $$current[$k];
			
			if ($info eq "Scenario")
			{
				if (scalar keys %scenarios ne 0)
				{
					$gFtsScenarios{$scenarios{"id"}} = {%scenarios};
					%scenarios = ();
				}
			}
			
			if (($info eq "id") or ($info eq "headline") or ($info eq "body"))
			{
				$tag = trim($info);
			}
			elsif (($tag ne "") and ($info ne 0) and (trim($info) ne ""))
			{
				$scenarios{$tag} = trim($info);
			}
		}
	}
	
	if ((scalar keys %scenarios ne 0) and ($tag eq ""))
	{
		if (!(exists $gFtsScenarios{$scenarios{"id"}}))
		{
			# the last element was not pushed into the hash
			$gFtsScenarios{$scenarios{"id"}} = {%scenarios};
			%scenarios = ();
		}
	}
}


##############################################################################
# Finds out if the default 'cqsearchprofile' can be deleted or not
sub canDeleteDefaultProfile_cqsearchprofile
{
	my $deleteCqsearchprofile = $ENV{"CQ_CAN_DELETE_CQSEARCHPROFILE"};
	if (defined($deleteCqsearchprofile))
	{
		$deleteCqsearchprofile = 1;
	}
	else
	{
		$deleteCqsearchprofile = 0;
	}

	return ($deleteCqsearchprofile);
}


##############################################################################
# Sets default values which can be overridden by command-line option or the 
# configuration file
sub setDefaultValues
{
	if (!exists $defValues{ "destination" })
	{
		$defValues{ "destination" } = ($gIsWindows eq 1) ? "C:/CQ.Search" : "/CQ.Search";
	}
	if (!exists $defValues{ "dbSet" })
	{
		$defValues{ "dbSet" } = "TextSearch";
	}
	if (!exists $defValues{ "userDB" })
	{
		$defValues{ "userDB" } = "SAMPL";
	}
	if (!exists $defValues{ "siteName" })
	{
		$defValues{ "siteName" } = "&lt;local&gt;";
	}
	if (!exists $defValues{ "enableAtSites" })
	{
		$defValues{ "enableAtSites" } = "";
	}
	if (!exists $defValues{ "entityFile" })
	{
		$defValues{ "entityFile" } = "";
	}
	if (!exists $defValues{ "ftsWASProfileName" })
	{
		$defValues{ "ftsWASProfileName" } = "cqsearchprofile";
	}
	if (!exists $defValues{ "ftsWASProfilePort" })
	{
		$defValues{ "ftsWASProfilePort" } = "14080";
	}
	if (!exists $defValues{ "ftsServerName" })
	{
		$defValues{ "ftsServerName" } = "localhost";
	}
	if (!exists $defValues{ "enableCQFTS" })
	{
		$defValues{ "enableCQFTS" } = "true";
	}
	if (!exists $defValues{ "runBatchIndex" })
	{
		$defValues{ "runBatchIndex" } = "true";
	}
	if (!exists $defValues{ "startUpdateIndex" })
	{
		$defValues{ "startUpdateIndex" } = "true";
	}
	if (!exists $defValues{ "batchIndexJVMParm" })
	{
		$defValues{ "batchIndexJVMParm" } = "";
	}
	if (!exists $defValues{ "ftsWASProfileMaxHeapSize" })
	{
		$defValues{ "ftsWASProfileMaxHeapSize" } = "";
	}
	if (!exists $defValues{ "batchIndexBatchSize" })
	{
		$defValues{ "batchIndexBatchSize" } = "250";
	}
	if (!exists $defValues{ "batchIndexDelay" })
	{
		$defValues{ "batchIndexDelay" } = "30";
	}
	if (!exists $defValues{ "batchIndexThreads" })
	{
		$defValues{ "batchIndexThreads" } = "5";
	}
	if (!exists $defValues{ "updateIndexBatchSize" })
	{
		$defValues{ "updateIndexBatchSize" } = "250";
	}
	if (!exists $defValues{ "updateIndexDelay" })
	{
		$defValues{ "updateIndexDelay" } = "30";
	}
	if (!exists $defValues{ "updateIndexLoginInterval" })
	{
		$defValues{ "updateIndexLoginInterval" } = "1";
	}
	if (!exists $defValues{ "addRecordType" })
	{
		$defValues{ "addRecordType" } = "";
	}
	if (!exists $defValues{ "removeRecordType" })
	{
		$defValues{ "removeRecordType" } = "";
	}
	if (!exists $defValues{ "deleteFtsWASProfiles" })
	{
		$defValues{ "deleteFtsWASProfiles" } = "";
	}
	
	$ftsCQPerl = ($gIsWindows eq 1) ? $ftsClearQuestHomeDir."cqperl" : "cqperl";
}


##############################################################################
# Gets the user's value from the configuration file
sub getUserValue
{
	my ($key) = @_;
	
	my $value = $newValues{$key};
	
	if ((!defined $value) or (length($value) == 0))
	{
		$value = $defValues{$key};
	}
	
	if ((!defined $value) or (length($value) == 0))
	{
		return ""; # there is no value, so assume none; this maybe OK for some cases, but not all
	}
	
	return $value;
}


##############################################################################
# Navigates between directories
sub followDir
{
	my ($path, $code_ref) = @_;
	my ($DH, $file);
	
	if (-d $path)
	{
		$code_ref->($path);
        
		unless (opendir $DH, $path)
		{
			dealWithError($langMsgText->maketext("CANNOT_OPEN", $path)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
        
		while ($file = readdir $DH)
		{
			next if $file eq '.' || $file eq '..';
			followDir("$path/$file", $code_ref);
		}
	}
	else
	{
		my $toWhere = $path;
		
		substr($toWhere, 0, length($ftsSourceDir), $ftsTargetDir);
		if (!copy($path, $toWhere))
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_FROM_TO", $path, $toWhere)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
}


##############################################################################
# Copies a directory (and it's content)
sub copyDir
{
	my ($copy) = @_;
	
	substr($copy, 0, length($ftsSourceDir), $ftsTargetDir);
	
	unless (-d $copy)
	{
		if (mkdir $copy)
		{
			#printAndLog "Created $copy"."\n";
		}
		else
		{
			dealWithError($langMsgText->maketext("CANNOT_CREATE", $copy)."\n".$msgLable_oS_ERROR.$!."\n", 1);
		}
	}
	else
	{
		if ($copy ne $ftsDataDir)
		{
			dealWithError($langMsgText->maketext("CANNOT_COPY_ALREADY_EXIST", $copy, $ftsTargetDir)."\n", 1);
		}
	}
}


##############################################################################
# Handles an element start event of XML data parsing
sub startElement
{
	my ($parseinst, $element, %attrs) = @_;
	
	SWITCH:
	{
		if ($element eq "destination")
		{
			$tag = "destination";
			last SWITCH;
		}
		if ($element eq "dbSet")
		{
			$tag = "dbSet";
			last SWITCH;
		}
		if ($element eq "userDB")
		{
			$tag = "userDB";
			last SWITCH;
		}
		if ($element eq "siteName")
		{
			$tag = "siteName";
			last SWITCH;
		}
		if ($element eq "enableAtSites")
		{
			$tag = "enableAtSites";
			last SWITCH;
		}
		if ($element eq "entityFile")
		{
			$tag = "entityFile";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileName")
		{
			$tag = "ftsWASProfileName";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePort")
		{
			$tag = "ftsWASProfilePort";
			last SWITCH;
		}
		if ($element eq "ftsServerName")
		{
			$tag = "ftsServerName";
			last SWITCH;
		}
		if ($element eq "enableCQFTS")
		{
			$tag = "enableCQFTS";
			last SWITCH;
		}
		if ($element eq "runBatchIndex")
		{
			$tag = "runBatchIndex";
			last SWITCH;
		}
		if ($element eq "startUpdateIndex")
		{
			$tag = "startUpdateIndex";
			last SWITCH;
		}
		if ($element eq "batchIndexJVMParm")
		{
			$tag = "batchIndexJVMParm";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileMaxHeapSize")
		{
			$tag = "ftsWASProfileMaxHeapSize";
			last SWITCH;
		}
		if ($element eq "batchIndexBatchSize")
		{
			$tag = "batchIndexBatchSize";
			last SWITCH;
		}
		if ($element eq "batchIndexDelay")
		{
			$tag = "batchIndexDelay";
			last SWITCH;
		}
		if ($element eq "batchIndexThreads")
		{
			$tag = "batchIndexThreads";
			last SWITCH;
		}
		if ($element eq "updateIndexBatchSize")
		{
			$tag = "updateIndexBatchSize";
			last SWITCH;
		}
		if ($element eq "updateIndexDelay")
		{
			$tag = "updateIndexDelay";
			last SWITCH;
		}
		if ($element eq "updateIndexLoginInterval")
		{
			$tag = "updateIndexLoginInterval";
			last SWITCH;
		}
		if ($element eq "addRecordType")
		{
			$tag = "addRecordType";
			last SWITCH;
		}
		if ($element eq "removeRecordType")
		{
			$tag = "removeRecordType";
			last SWITCH;
		}
		if ($element eq "deleteFtsWASProfiles")
		{
			$tag = "deleteFtsWASProfiles";
			last SWITCH;
		}
        
		if ($element eq "defaultValue")
		{
			$isInDefaultValue = 1;
			last SWITCH;
		}
		if ($element eq "newValue")
		{
			$isInNewValue = 1;
			last SWITCH;
		}
	}
}


##############################################################################
# Handles an element end event of XML data parsing
sub endElement
{
	my ($parseinst, $element, %attrs) = @_;
	
	SWITCH:
	{
		if ($element eq "destination")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "dbSet")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "userDB")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "siteName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "enableAtSites")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "entityFile")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfilePort")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsServerName")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "enableCQFTS")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "runBatchIndex")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "startUpdateIndex")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexJVMParm")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "ftsWASProfileMaxHeapSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexBatchSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexDelay")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "batchIndexThreads")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "updateIndexBatchSize")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "UpdateIndexDelay")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "updateIndexLoginInterval")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "addRecordType")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "removeRecordType")
		{
			$tag = "";
			last SWITCH;
		}
		if ($element eq "deleteFtsWASProfiles")
		{
			$tag = "";
			last SWITCH;
		}
		
		if ($element eq "defaultValue")
		{
			$isInDefaultValue = 0;
			$defaultValue = "";
			last SWITCH;
		}
		if ($element eq "newValue")
		{
			$isInNewValue = 0;
			$newValue = "";
			last SWITCH;
		}
	}
}


##############################################################################
# Handles character data event of XML data parsing
sub characterData
{
	my ($parseinst, $data) = @_;
	
	if ($isInDefaultValue == 1)
	{
		$defValues{ $tag } .= trim(HTML::Entities::decode($data));
	}
	elsif ($isInNewValue == 1)
	{
		$newValues{ $tag } .= trim(HTML::Entities::decode($data));
	}
}


##############################################################################
# Removes white characters from start and end of a string
sub trim
{
	my ($string) = @_;
	
	if (defined($string))
	{
		$string =~ s/^\s+//;
		$string =~ s/\s+$//;
	}
	
	return $string;
}


##############################################################################
# Compares the data of 2 arrays
sub compareArrays
{
	my ($a1, $a2) = @_;
	
	my @aa1 = sort(split(",", $a1));
	my @aa2 = sort(split(",", $a2));
	
	return 0 unless @aa1 == @aa2;
	
	for (my $i = 0; $i < @aa1; $i++)
	{
		return 0 if trim($aa1[$i]) ne trim($aa2[$i]);
	}
	return 1;
}


##############################################################################
# Clears an array list of Entity data
sub cleanArrayList
{
	my ($s) = @_;
	
	my @a = split(",", $s);
	my $result = "";
	
	@a = sort(@a);
	
	foreach my $i (@a)
	{
		if ($result ne "")
		{
			$result .= ",";
		}
		$result .= trim($i);
	}
	
	return $result;
}


##############################################################################
# For debugging, prints the key=value paris
sub printValues
{
	for my $key ( keys %defValues )
	{
		my $value = $defValues{$key};
		
		printAndLog "Key =======> ".$key."\n";
		printAndLog "val [def] => ".$value."\n";
		printAndLog "val [new] => ".$newValues{$key}."\n";
	}
}


##############################################################################
# Prints the small help for the tool
sub printMiniUsage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printAndLog $langMsgText->maketext("MINI_HELP_MSG", "1.0.0.04-D120518"), 1;
	
	if ($doExit)
	{
		exit $doExit
	}
}


##############################################################################
# Prints the long help for the tool
sub printLongUsage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printAndLog $langMsgText->maketext("FULL_HELP_MSG"), 1;
	
	if ($doExit)
	{
		exit $doExit
	}
}


##############################################################################
# Prints the usage for the tool
sub Usage
{
	my ($dieMsg, $doExit) = @_;
	
	printAndLog $dieMsg, 1;
	
	printMiniUsage("", 0);
	printLongUsage("", 0);
	
	if ($doExit)
	{
		exit $doExit
	}
}
